# Copyright (c) 2025 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

import pytest
import numpy as np
import gen_tso.catalogs as cat
from gen_tso.catalogs.fetch_catalogs import (
    fetch_simbad_aliases,
    fetch_aliases,
)


@pytest.mark.parametrize(
    'target',
    [
        'WASP-69',
        ['WASP-69'],
    ],
)
def test_fetch_simbad_aliases_single(target):
    aliases, ks_mag = fetch_simbad_aliases(target)

    expected_aliases = [
        'AP J21000618-0505398',
        'NAME Wouri',
        'Gaia DR3 6910753016653587840',
        'TIC 248853232',
        'BD-05  5432',
        'GSC 05200-01560',
        'PPM 204700',
        'TYC 5200-1560-1',
        'Gaia DR2 6910753016653587840',
        '2MASS J21000618-0505398',
        'WASP-69',
        'Gaia DR1 6910753012357203968',
    ]
    assert len(aliases[0]) == len(expected_aliases)
    assert aliases[0] == expected_aliases
    np.testing.assert_allclose(ks_mag, 7.45900011)


def test_fetch_simbad_aliases_planets():
    # Simbad thinks the TrES targets are the planets
    targets = ['TrES-4']
    aliases, ks_mag = fetch_simbad_aliases(targets)
    expected_aliases = [
        'TOI-2124',
        'TOI-2124',
        'TrES-4',
        'HIDDEN NAME TrES-4',
    ]
    assert len(aliases[0]) == len(expected_aliases)
    assert aliases[0] == expected_aliases
    assert np.isnan(ks_mag[0])


def test_fetch_simbad_aliases_not_found():
    aliases, ks_mag = fetch_simbad_aliases(['WASP-00'])
    assert aliases[0] == []
    assert np.isnan(ks_mag[0])


# public
def test_fetch_nea_aliases_single():
    targets = 'KELT-7'
    host_aliases, planet_aliases = cat.fetch_nea_aliases(targets)
    assert planet_aliases == [{
        'HD 33643 b': 'KELT-7 b',
        'KELT-7 b': 'KELT-7 b',
        'BD+33 977 b': 'KELT-7 b',
        'HIP 24323 b': 'KELT-7 b',
        'SAO 57753 b': 'KELT-7 b',
        'GSC 2393-00852 b': 'KELT-7 b',
        'TYC 2393-00852-1 b': 'KELT-7 b',
        '2MASS J05131092+3319054 b': 'KELT-7 b',
        'WISE J051310.93+331904.8 b': 'KELT-7 b',
        'TIC 367366318 b': 'KELT-7 b',
        'Gaia DR2 181908842994567936 b': 'KELT-7 b',
        'TOI-1682.01': 'KELT-7 b',
        'TOI-1682 b': 'KELT-7 b',
    }]
    assert host_aliases == [{
        'TIC 367366318': 'KELT-7',
        'Gaia DR2 181908842994567936': 'KELT-7',
        'TOI-1682': 'KELT-7',
        'HD 33643': 'KELT-7',
        'KELT-7': 'KELT-7',
        'BD+33 977': 'KELT-7',
        'HIP 24323': 'KELT-7',
        'SAO 57753': 'KELT-7',
        'GSC 2393-00852': 'KELT-7',
        'TYC 2393-00852-1': 'KELT-7',
        '2MASS J05131092+3319054': 'KELT-7',
        'WISE J051310.93+331904.8': 'KELT-7',
    }]


def test_fetch_nea_aliases_basics():
    targets = ['WASP-8 b', 'KELT-7']
    host_aliases, planet_aliases = cat.fetch_nea_aliases(targets)

    assert len(host_aliases) == 2
    assert len(planet_aliases) == 2

    expected_host_aliases = {
        '1SWASP J235936.07-350152.9',
        '2MASS J23593607-3501530',
        'Gaia DR2 2312679845530628096',
        'TIC 183532609',
        'TOI-191',
        'TYC 7522-00505-1',
        'WASP-8',
        'WASP-8 A',
        'WISE J235936.16-350153.1',
    }
    assert set(host_aliases[0]) == expected_host_aliases
    assert np.unique(list(host_aliases[0].values())) == 'WASP-8'
    assert np.unique(list(host_aliases[1].values())) == 'KELT-7'

    expected_planets_wasp = set(planet_aliases[0].values())
    expected_planets_kelt = set(planet_aliases[1].values())
    assert expected_planets_wasp == {'WASP-8 b', 'WASP-8 c'}
    assert expected_planets_kelt == {'KELT-7 b'}


def test_fetch_nea_aliases_multi_systems():
    targets = [
        'WASP-8',   # 2x star, only one has two planets
        'TOI-1338', # 2x star, 2 planets transiting both stars
        'WASP-94',  # 2x star, 1 planet each
    ]
    host_aliases, planet_aliases = cat.fetch_nea_aliases(targets)

    assert len(host_aliases) == 3
    assert len(planet_aliases) == 3

    # Check identifiers
    expected_hosts = [
        set(['WASP-8']),
        set(['TOI-1338 A', 'TOI-1338 B']),
        set(['WASP-94 A', 'WASP-94 B']),
    ]
    assert set(host_aliases[0].values()) == expected_hosts[0]
    assert set(host_aliases[1].values()) == expected_hosts[1]
    assert set(host_aliases[2].values()) == expected_hosts[2]

    expected_planets = [
        set(['WASP-8 b', 'WASP-8 c']),
        set(['TOI-1338 b', 'TOI-1338 c']),
        set(['WASP-94 A b', 'WASP-94 B b']),
    ]
    assert set(planet_aliases[0].values()) == expected_planets[0]
    assert set(planet_aliases[1].values()) == expected_planets[1]
    assert set(planet_aliases[2].values()) == expected_planets[2]

    # Check planet aliases WASP-8:
    expected_aliases_WASP8b = set([
        'WASP-8 b',
        'TYC 7522-00505-1 b',
        '2MASS J23593607-3501530 b',
        'WISE J235936.16-350153.1  b',
        '1SWASP J235936.07-350152.9 b',
        'TIC 183532609 b',
        'Gaia DR2 2312679845530628096 b',
        'TOI-191.01',
        'TOI-191 b',
    ])
    expected_aliases_WASP8c = set([
        'TIC 183532609 c',
        'Gaia DR2 2312679845530628096 c',
        'WASP-8 c',
        'WASP-8 A b',
        'WASP-8 A c',
        'TYC 7522-00505-1 c',
        '2MASS J23593607-3501530 c',
        'WISE J235936.16-350153.1  c',
        '1SWASP J235936.07-350152.9 c',
    ])
    aliases_WASP8b = set([
        key for key,val in planet_aliases[0].items()
        if val == 'WASP-8 b'
    ])
    aliases_WASP8c = set([
        key for key,val in planet_aliases[0].items()
        if val == 'WASP-8 c'
    ])
    assert aliases_WASP8b == expected_aliases_WASP8b
    assert aliases_WASP8c == expected_aliases_WASP8c

    # Check planet aliases WASP-94:
    expected_aliases_WASP94Ab = set([
        'WASP-94 A b',
        'TYC 7466-1400-1 b',
        '2MASS J20550794-3408079 b',
        'WISE J205507.96-340808.4 b',
        '1SWASP J205507.94-340807.9 A b',
        'TIC 92352620 b',
        'Gaia DR2 6780546169633475456 b',
        'TOI-107.01',
        'TOI-107 b',
    ])
    expected_aliases_WASP94Bb = set([
        'WASP-94 B b',
        '2MASS J20550915-3408078 b',
        'WISE J205509.17-340808.3 b',
        '1SWASP J205507.94-340807.9 B b',
        'TIC 92352621 b',
        'Gaia DR2 6780546169633474944 b',
    ])
    aliases_WASP94Ab = set([
        key for key,val in planet_aliases[2].items()
        if val == 'WASP-94 A b'
    ])
    aliases_WASP94Bb = set([
        key for key,val in planet_aliases[2].items()
        if val == 'WASP-94 B b'
    ])
    assert aliases_WASP94Ab == expected_aliases_WASP94Ab
    assert aliases_WASP94Bb == expected_aliases_WASP94Bb


def test_fetch_nea_aliases_oddballs():
    hosts = [
        'TOI-4336',
        'TOI-216',
        'WASP-50',
        'WASP-53',
        'WASP-76',
        '55 Cnc',
    ]
    host_aliases, planet_aliases = cat.fetch_nea_aliases(hosts)



def test_fetch_nea_aliases_not_found():
    targets = 'WASP-999'
    host_aliases, planet_aliases = cat.fetch_nea_aliases(targets)
    assert host_aliases == [{}]
    assert planet_aliases == [{}]



def test_fetch_aliases_basics():
    hosts = [
        'WASP-69',
        'TOI-270',
    ]
    aliases = fetch_aliases(hosts)

    assert list(aliases) == hosts

    expected_keys = ['host', 'planets', 'host_aliases', 'planet_aliases']
    for key in expected_keys:
        assert key in aliases[hosts[0]].keys()

    expected_toi_planets = ['TOI-270 b', 'TOI-270 c', 'TOI-270 d']
    assert aliases[hosts[0]]['planets'] == ['WASP-69 b']
    assert list(aliases[hosts[1]]['planets']) == expected_toi_planets

    expected_planet_aliases = {
        'WASP-69 b': 'WASP-69 b',
        'BD-05 5432 b': 'WASP-69 b',
        'GSC 05200-01560 b': 'WASP-69 b',
        'TYC 5200-1560-1 b': 'WASP-69 b',
        '2MASS J21000618-0505398 b': 'WASP-69 b',
        'WISE J210006.21-050540.9 b': 'WASP-69 b',
        'TIC 248853232 b': 'WASP-69 b',
        'Gaia DR2 6910753016653587840 b': 'WASP-69 b',
        'TOI-5823.01': 'WASP-69 b',
        'TOI-5823 b': 'WASP-69 b',
        'Wouri b': 'WASP-69 b',
    }
    assert aliases[hosts[0]]['planet_aliases'] == expected_planet_aliases


def test_fetch_aliases_multiplanet_systems():
    hosts = [
        'WASP-8',   # 2x star, only one has two planets
        'TOI-1201',  # 2x star, 1 planet each
        'TOI-1338 A', # 2x star, 2 planets transiting both stars
    ]
    aliases = fetch_aliases(hosts)

    assert hosts[0] in aliases
    assert hosts[1] in aliases
    assert hosts[2] in aliases

    assert set(aliases[hosts[0]]['planets']) == {'WASP-8 b', 'WASP-8 c'}
    assert set(aliases[hosts[1]]['planets']) == {'TOI-1201 b'}
    assert set(aliases[hosts[2]]['planets']) == {'TOI-1338 b', 'TOI-1338 c'}

    host_aliases_WASP8 = {
        '1SWASP J235936.07-350152.9',
        '2MASS J23593607-3501530',
        'Gaia DR2 2312679845530628096',
        'TIC 183532609',
        'TOI-191',
        'TYC 7522-00505-1',
        'WASP-8',
        'WASP-8 A',
        'WISE J235936.16-350153.1',
    }
    host_aliases_TOI1201 = {
        '2MASS J02485926-1432152',
        'Gaia DR2 5157183324996790272',
        'TIC 29960110',
        'TOI-1201',
    }
    host_aliases_TOI1338 = {
        '2MASS J06083197-5932280',
        'BEBOP-1',
        'EBLM J0608-59',
        'Gaia DR2 5494443978353833088',
        'TIC 260128333',
        'TOI-1338 A',
        'TOI-1338 Aa',
        'TYC 8533-950-1',
        'WISE J060831.94-593227.6',
    }
    assert set(aliases[hosts[0]]['host_aliases']) == host_aliases_WASP8
    assert set(aliases[hosts[1]]['host_aliases']) == host_aliases_TOI1201
    assert set(aliases[hosts[2]]['host_aliases']) == host_aliases_TOI1338

    planet_aliases_WASP8 = {
        'WASP-8 b': 'WASP-8 b',
        'TYC 7522-00505-1 b': 'WASP-8 b',
        '2MASS J23593607-3501530 b': 'WASP-8 b',
        'WISE J235936.16-350153.1 b': 'WASP-8 b',
        '1SWASP J235936.07-350152.9 b': 'WASP-8 b',
        'TIC 183532609 b': 'WASP-8 b',
        'Gaia DR2 2312679845530628096 b': 'WASP-8 b',
        'TOI-191.01': 'WASP-8 b',
        'TOI-191 b': 'WASP-8 b',
        'TIC 183532609 c': 'WASP-8 c',
        'Gaia DR2 2312679845530628096 c': 'WASP-8 c',
        'WASP-8 c': 'WASP-8 c',
        'WASP-8 A b': 'WASP-8 b',
        'WASP-8 A c': 'WASP-8 c',
        'TYC 7522-00505-1 c': 'WASP-8 c',
        '2MASS J23593607-3501530 c': 'WASP-8 c',
        'WISE J235936.16-350153.1 c': 'WASP-8 c',
        '1SWASP J235936.07-350152.9 c': 'WASP-8 c',
    }
    planet_aliases_TOI1201 = {
        'TOI-1201 b': 'TOI-1201 b',
        'TOI-1201.01': 'TOI-1201 b',
        'TIC 29960110 b': 'TOI-1201 b',
        'TIC 29960110.01': 'TOI-1201 b',
        '2MASS J02485926-1432152 b': 'TOI-1201 b',
        'Gaia DR2 5157183324996790272 b': 'TOI-1201 b',
    }
    planet_aliases_TOI1338 = {
        'Gaia DR2 5494443978353833088 b': 'TOI-1338 b',
        'TOI-1338 b': 'TOI-1338 b',
        'TOI-1338.01': 'TOI-1338 b',
        'EBLM J0608-59 b': 'TOI-1338 b',
        'TIC 260128333 b': 'TOI-1338 b',
        'TIC 260128333.01': 'TOI-1338 b',
        'TYC 8533-950-1 b': 'TOI-1338 b',
        '2MASS J06083197-5932280 b': 'TOI-1338 b',
        'WISE J060831.94-593227.6 b': 'TOI-1338 b',
        'TOI-1338 c': 'TOI-1338 c',
        'TIC 260128333 c': 'TOI-1338 c',
        '2MASS J06083197-5932280 c': 'TOI-1338 c',
        'WISE J060831.94-593227.6 c': 'TOI-1338 c',
        'Gaia DR2 5494443978353833088 c': 'TOI-1338 c',
        'TYC 8533-950-1 c': 'TOI-1338 c',
        'EBLM J0608-59 c': 'TOI-1338 c',
        'BEBOP-1 c': 'TOI-1338 c',
        'BEBOP-1 b': 'TOI-1338 b',
    }
    assert aliases[hosts[0]]['planet_aliases'] == planet_aliases_WASP8
    assert aliases[hosts[1]]['planet_aliases'] == planet_aliases_TOI1201
    assert aliases[hosts[2]]['planet_aliases'] == planet_aliases_TOI1338


def test_fetch_aliases_name_change():
    hosts = [
        'TOI-1241',  # KOI-5
        'TOI-2410',  # EPIC 220198551
        'TOI-4454',  # Kepler-488
        'TOI-4588',  # HIP 92247
        'TOI-7383',  # Gaia-1
    ]
    aliases = fetch_aliases(hosts)

    expected_hosts = [
        'KOI-5',
        'EPIC 220198551',
        'Kepler-488',
        'HIP 92247',
        'Gaia-1',
    ]
    assert list(aliases) == expected_hosts
    for i in range(len(hosts)):
        host = expected_hosts[i]
        host_aliases = aliases[host]['host_aliases']
        assert hosts[i] in host_aliases


def test_fetch_gaia_targets():
    target_ra = 315.0260
    target_dec = -5.0949
    names, G_mag, teff, log_g, ra, dec, separation = cat.fetch_gaia_targets(
        target_ra, target_dec, max_separation=80.0,
    )
    expected_names = [
        'Gaia DR3 6910753016653587840', 'Gaia DR3 6910752844854895360',
        'Gaia DR3 6910747136843460480', 'Gaia DR3 6910746934979897088',
        'Gaia DR3 6910747141138328064', 'Gaia DR3 6910753046718453248',
        'Gaia DR3 6910746930684973952', 'Gaia DR3 6910746866260418944',
        'Gaia DR3 6910746934979895936',
    ]
    expected_G_mag = np.array([
         9.491504, 16.80548 , 18.476078, 16.334204, 18.974197, 17.175997,
        18.024723, 15.700529, 16.899416,
    ])
    expected_teff = np.array([
        4729.3774, 5503.5996, 4959.2495, 5394.8145, 3979.5588, 5134.7896,
        4528.2563, 5572.225 , 4862.2915,
    ])
    expected_log_g = np.array([
        4.5058, 4.4191, 4.7718, 4.3786, 5.018 , 4.3886, 4.6305, 3.774 ,
        4.6665,
    ])
    expected_ra = np.array([
        315.02597081, 315.03419028, 315.01496413, 315.01522348,
        315.00966205, 315.04041427, 315.02272526, 315.03525381,
        315.0218094,
    ])
    expected_dec = np.array([
        -5.09487006, -5.08816924, -5.09282223, -5.10531022, -5.10264537,
        -5.08275871, -5.11378437, -5.11325095, -5.11603216,
    ])
    expected_sep = np.array([
        0.15023734, 38.07422832, 40.27294104, 53.83032696, 64.88104462,
       67.69030447, 68.9903496 , 73.92831233, 77.54554983,
    ])
    assert names.tolist() == expected_names
    np.testing.assert_allclose(G_mag, expected_G_mag)
    np.testing.assert_allclose(teff, expected_teff)
    np.testing.assert_allclose(log_g, expected_log_g)
    np.testing.assert_allclose(ra, expected_ra)
    np.testing.assert_allclose(dec, expected_dec)
    np.testing.assert_allclose(separation, expected_sep)


@pytest.mark.skip(reason='mock requests')
def test_fetch_gaia_targets_error():
    pass




# Copyright (c) 2025 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

import os
import sys
from shiny import run_app



def main():
    """
    Launch the Gen TSO application.

    Usage
    -----
    # Run the application
    tso [--debug] [models_folder]

    # Check/update the pandeia reference data
    tso --update_db

    # Update the NASA Exoplanet Archive data
    tso --update_exo

    Optional commands
    -----------------
    --update_db:
        If set, update the pysynphot database
    --update_exo:
        If set, update the NASA exoplanet archive
    --debug:
        If set, run the app with reload=True, which reloads a live
        app if the code is updated.
    models_folder:
        If set, the app will attempt to load transit, eclipse, and SED
        models from the specified folder.
    """
    if '--update_db' in sys.argv:
        # Import here, otherwise shiny breaks (bc of grequests?)
        from gen_tso.pandeia_io.pandeia_setup import update_synphot_files
        status = update_synphot_files()

    if '--update_exo' in sys.argv:
        import gen_tso.catalogs as cat
        cat.update_exoplanet_archive()

    if '--update_programs' in sys.argv:
        import gen_tso.catalogs as cat
        cat.update_jwst_programs()

    if (
        '--update_db' not in sys.argv and
        '--update_exo' not in sys.argv and
        '--update_programs' not in sys.argv
    ):
        reload = '--debug' in sys.argv
        app = os.path.realpath(os.path.dirname(__file__)) + '/gen_tso_app.py'
        run_app(app, reload=reload, launch_browser=True, dev_mode=False)


if __name__ == "__main__":
    main()




# Copyright (c) 2025 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

import pathlib
import numpy as np
from gen_tso.utils import ROOT
import gen_tso.catalogs.utils as u

# Path to tests folder
ROOT = pathlib.Path(ROOT).parent.joinpath('tests')


def test_esasky_js_circle():
    ra = 315.0259661
    dec = -5.094857
    radius = 80.0
    circle = u.esasky_js_circle(ra, dec, radius)


    expected_circle = {
        'event': 'overlayFootprints',
        'content': {
            'overlaySet': {
                'type': 'FootprintListOverlay',
                'overlayName': 'visit splitting distance',
                'cooframe': 'J2000',
                'color': '#15B01A',
                'lineWidth': 5,
                'skyObjectList': [{
                     'name': 'visit splitting distance',
                     'id': 1,
                     'stcs': 'CIRCLE ICRS 315.02596610 -5.09485700 0.0222',
                     'ra_deg': '315.02596610',
                     'dec_deg': '-5.09485700',
                }]
            }
        }
    }
    assert circle == expected_circle


def test_json_target_property():
    name = 'G mag'
    value = 9.491504
    format = '.2f'
    prop = u.json_target_property(name, value, format)
    expected_prop = {'name': 'G mag', 'value': '9.49', 'type': 'STRING'}
    assert prop == expected_prop


def test_json_target():
    # e.g., from a cat.fetch_gaia_targets(ra_source, dec_source) call
    i = 0
    name = 'Gaia DR3 6910753016653587840'
    g_mag = 9.491504
    teff = 4729.3774
    logg = 4.5058
    ra = 315.02597081
    dec = -5.09487006
    separation = 0.15023734
    j_target = u.json_target(i, name, ra, dec, g_mag, teff, logg, separation)

    expected_target = {
        'name': 'Gaia DR3 6910753016653587840',
        'id': 1,
        'ra': '315.02597081',
        'dec': '-5.09487006',
        'data': [
            {'name': 'G mag', 'value': '9.49', 'type': 'STRING'},
            {'name': 'T eff', 'value': '4729.4', 'type': 'STRING'},
            {'name': 'log(g)', 'value': '4.51', 'type': 'STRING'},
            {'name': 'Separation', 'value': '0.150', 'type': 'STRING'}
        ]
    }
    assert j_target == expected_target


def test_esasky_js_catalog():
    pass


def test_normalize_name_odd():
    assert u.normalize_name('-RHO01-CNC') == '55 Cnc'
    assert u.normalize_name('-rho01-Cnc') == '55 Cnc'
    assert u.normalize_name('55CNC') == '55 Cnc'

    assert u.normalize_name('NAME-G-268-38B') == 'G 268-38'

    assert u.normalize_name('WD-1856+534-B') == 'WD 1856+534'
    assert u.normalize_name('WD1856+534') == 'WD 1856+534'
    assert u.normalize_name('WD1856B') == 'WD 1856+534'

    assert u.normalize_name('V-V1298-TAU') == 'V1298 Tau'
    assert u.normalize_name('V-V1298-Tau') == 'V1298 Tau'


def test_normalize_name_tails():
    assert u.normalize_name('CD-38-1467-copy') == 'CD-38 1467'
    assert u.normalize_name('GJ-436-offset') ==  'GJ 436'
    assert u.normalize_name('L98-59-updated') == 'L 98-59'
    assert u.normalize_name('TOI-455-revised') == 'TOI-455'


def test_normalize_name_case():
    assert u.normalize_name('KEPLER-12') == 'Kepler-12'
    assert u.normalize_name('KEPLER-86B') == 'Kepler-86'
    assert u.normalize_name('Kepler-12') == 'Kepler-12'
    assert u.normalize_name('TRES-4') == 'TrES-4'
    assert u.normalize_name('WOLF-437') == 'Wolf 437'


def test_normalize_name_BD():
    assert u.normalize_name('BD+01-316') == 'BD+01 316'
    assert u.normalize_name('BD-17-588A') == 'BD-17 588 A'


def test_normalize_name_HATs():
    assert u.normalize_name('HAT-P-1') == 'HAT-P-1'
    assert u.normalize_name('HAT-P-26B') == 'HAT-P-26'
    assert u.normalize_name('HATP1') == 'HAT-P-1'
    assert u.normalize_name('HATS-6') == 'HATS-6'
    assert u.normalize_name('HATS-72b') == 'HATS-72'


def test_normalize_name_HDs():
    assert u.normalize_name('HD-12572b') == 'HD 12572'
    assert u.normalize_name('HD-133112') == 'HD 133112'
    assert u.normalize_name('HD-189733B') == 'HD 189733'
    assert u.normalize_name('HD106315') == 'HD 106315'


def test_normalize_name_GJs():
    assert u.normalize_name('GJ-1132') == 'GJ 1132'
    assert u.normalize_name('GJ-4102A') == 'GJ 4102 A'
    assert u.normalize_name('GJ1132') == 'GJ 1132'
    assert u.normalize_name('GL486') == 'GJ 486'


def test_normalize_name_Ls():
    assert u.normalize_name('L-231-32') == 'L 231-32'
    assert u.normalize_name('L168-9') == 'L 168-9'

    assert u.normalize_name('LHS-1140') == 'LHS 1140'
    assert u.normalize_name('LHS1140') == 'LHS 1140'
    assert u.normalize_name('LHS_3844') == 'LHS 3844'

    assert u.normalize_name('LP-141-14') == 'LP 141-14'

    assert u.normalize_name('LTT-1445A') == 'LTT 1445 A'
    assert u.normalize_name('LTT-3780') == 'LTT 3780'
    assert u.normalize_name('LTT1445A') == 'LTT 1445 A'
    assert u.normalize_name('LTT9779') == 'LTT 9779'


def test_normalize_name_TOIs():
    assert u.normalize_name('TOI-1075') == 'TOI-1075'
    assert u.normalize_name('TOI-1807b') == 'TOI-1807'


def test_normalize_name_K2s():
    assert u.normalize_name('K2-141') == 'K2-141'
    assert u.normalize_name('K2-22B') == 'K2-22'


def test_normalize_name_WASPs():
    assert u.normalize_name('WASP-103') == 'WASP-103'
    assert u.normalize_name('WASP-121B') == 'WASP-121'
    assert u.normalize_name('WASP-39b') == 'WASP-39'
    assert u.normalize_name('WASP-77A') == 'WASP-77 A'


def test_normalize_name_others():
    assert u.normalize_name('2MASS-J11335277+1227034') == '2MASS J11335277+1227034'
    assert u.normalize_name('AU-MIC') == 'AU Mic'
    assert u.normalize_name('CD-38-1467') == 'CD-38 1467'
    assert u.normalize_name('HIP67522') == 'HIP 67522'
    assert u.normalize_name('IRAS-04125+2902') == 'IRAS 04125+2902'
    assert u.normalize_name('KELT-20') == 'KELT-20'
    assert u.normalize_name('NGTS-10') == 'NGTS-10'
    assert u.normalize_name('TRAPPIST-1') == 'TRAPPIST-1'
    assert u.normalize_name('TRAPPIST-1B') == 'TRAPPIST-1'
    assert u.normalize_name('TYC-7052-1753-1') == 'TYC 7052-1753-1'


def test_is_letter_yes():
    assert u.is_letter('WASP-69 b') is True


def test_is_letter_no():
    assert u.is_letter('TOI-741.01') is False
    assert u.is_letter('WASP-69') is False
    # TBD: check
    assert u.is_letter('WASP-8 A') is False


def test_is_candidate_yes():
    assert u.is_candidate('TOI-741.01') is True


def test_is_candidate_no():
    assert u.is_candidate('WASP-69 b') is False
    assert u.is_candidate('WASP-69') is False
    assert u.is_candidate('WASP-8 A') is False


def test_get_letter_ok():
    assert u.get_letter('WASP-69 b') == ' b'
    assert u.get_letter('TOI-741.01') == '.01'

def test_get_letter_fail():
    assert u.get_letter('WASP-69') == ''
    assert u.get_letter('WASP-8 A') == ''

def test_get_host_ok():
    assert u.get_host('WASP-69 b') == 'WASP-69'
    assert u.get_host('TOI-741.01') == 'TOI-741'


def test_get_host_fail():
    assert u.get_host('WASP-69') == ''
    # TBD: is this a problem?
    assert u.get_host('WASP-8 A') == ''


def test_to_float_float_int():
    value = u.to_float('101')
    np.testing.assert_equal(value, 101.0)


def test_to_float_float():
    value = u.to_float('1.01')
    np.testing.assert_equal(value, 1.01)


def test_to_float_exp():
    value = u.to_float('3.0e3')
    np.testing.assert_equal(value, 3000.0)


def test_to_float_none():
    assert u.to_float('None') is None


def as_str_ok():
    string = u.as_str(1.0)
    assert string == '1.0'


def as_str_exp():
    string = u.as_str(3.0e2)
    assert string == '300.0'


def as_str_nan():
    string = u.as_str(np.nan)
    assert string is None


def as_str_my_nan():
    string = u.as_str(np.nan, if_none='---')
    assert string == '---'


def as_str_fmt():
    string = u.as_str(1.0, fmt='.3f')
    assert string == '1.000'





# Copyright (c) 2025 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

import gen_tso.pandeia_io as jwst

import numpy as np
import pytest
import re



def test_get_instruments():
    instruments = jwst.get_instruments()

    expected_instruments = [
        'miri',
        'nircam',
        'niriss',
        'nirspec',
    ]
    assert instruments == expected_instruments


def test_get_modes_miri():
    modes = jwst.get_modes('miri')

    expected_modes = [
        'lrsslitless',
        'lrsslit',
        'mrs_ts',
        'imaging_ts',
        'target_acq',
    ]
    assert modes == expected_modes


def test_get_modes_nircam():
    modes = jwst.get_modes('nircam')

    expected_modes = [
        'lw_tsgrism',
        'sw_tsgrism',
        'lw_ts',
        'sw_ts',
        'target_acq',
    ]
    assert modes == expected_modes


def test_get_modes_nirspec():
    modes = jwst.get_modes('nirspec')

    expected_modes = [
        'bots',
        'target_acq',
    ]
    assert modes == expected_modes


def test_get_modes_niriss():
    modes = jwst.get_modes('niriss')

    expected_modes = [
        'soss',
        'target_acq',
    ]
    assert modes == expected_modes


def test_get_modes_type_spectro():
    modes = jwst.get_modes('miri', type='spectroscopy')
    expected_modes = [
        'lrsslitless',
        'lrsslit',
        'mrs_ts',
    ]
    assert modes == expected_modes


def test_get_modes_type_photo():
    modes = jwst.get_modes('miri', type='photometry')
    expected_modes = [
        'imaging_ts',
    ]
    assert modes == expected_modes


def test_get_modes_type_acq():
    expected_modes = ['target_acq']
    for inst in jwst.get_instruments():
        modes = jwst.get_modes(inst, type='acquisition')
        assert modes == expected_modes


def test_get_sed_types():
    sed_types = jwst.get_sed_types()
    expected_sed_types = [
        'phoenix',
        'k93models',
        'bt_settl',
    ]
    assert sed_types == expected_sed_types


def test_get_throughputs_all():
    throughputs = jwst.get_throughputs()

    assert list(throughputs) == ['spectroscopy', 'photometry', 'acquisition']

    expected_inst = ['miri', 'nircam', 'niriss', 'nirspec']
    assert list(throughputs['spectroscopy']) == expected_inst

    assert list(throughputs['spectroscopy']['nirspec']) == ['bots']

    subarrays = throughputs['spectroscopy']['nirspec']['bots']
    expected_subs = ['sub512', 'sub512s', 'sub1024a', 'sub1024b', 'sub2048']
    assert list(subarrays) == expected_subs

    filters = list(throughputs['spectroscopy']['nirspec']['bots']['sub2048'])
    expected_filters = [
        'g140m/f070lp',
        'g140h/f070lp',
        'g140m/f100lp',
        'g140h/f100lp',
        'g235m/f170lp', 
        'g235h/f170lp',
        'g395m/f290lp',
        'g395h/f290lp',
        'prism/clear',
    ]
    assert list(filters) == expected_filters


def test_get_throughputs_type():
    throughputs = jwst.get_throughputs(type='acquisition')
    assert list(throughputs) == ['miri', 'nircam', 'niriss', 'nirspec']
    assert list(throughputs['nirspec']) == ['target_acq']


def test_get_throughputs_inst():
    throughputs = jwst.get_throughputs(inst='nirspec')
    assert list(throughputs) == ['bots', 'target_acq']

    expected_subs = ['sub512', 'sub512s', 'sub1024a', 'sub1024b', 'sub2048']
    assert list(throughputs['bots']) == expected_subs


def test_get_throughputs_mode():
    throughputs = jwst.get_throughputs(inst='nirspec', mode='bots')
    assert list(throughputs) == ['sub512', 'sub512s', 'sub1024a', 'sub1024b', 'sub2048']


def test_get_throughputs_mode_missing_inst():
    error = 'Also need to specify an instrument when requesting a specific mode'
    with pytest.raises(ValueError, match=re.escape(error)):
        throughputs = jwst.get_throughputs(mode='bots')



def test_generate_all_instruments():
    detectors = jwst.generate_all_instruments()
    assert len(detectors) == 14
    # TBD: what to test?



def test_get_configs_miri_lrsslitless():
    configs = jwst._get_configs(instrument='miri', obs_type='spectroscopy')
    assert len(configs) == 3
    inst = configs[0]
    assert inst['mode'] == 'lrsslitless'
    assert inst['mode_label'] == 'Low Resolution Spectroscopy (LRS) Slitless'


def test_get_configs_miri_mrs_ts():
    configs = jwst._get_configs(instrument='miri', obs_type='spectroscopy')
    inst = configs[2]
    assert inst['mode'] == 'mrs_ts'
    assert inst['mode_label'] == 'MRS Time Series'


def test_get_configs_nircam_lw_tsgrism():
    configs = jwst._get_configs(instrument='nircam', obs_type='spectroscopy')
    inst = configs[0]
    assert inst['mode'] == 'lw_tsgrism'
    assert inst['mode_label'] == 'LW Grism Time Series'


def test_get_configs_nircam_sw_tsgrism():
    configs = jwst._get_configs(instrument='nircam', obs_type='spectroscopy')
    assert len(configs) == 2
    inst = configs[1]
    assert inst['mode'] == 'sw_tsgrism'
    assert inst['mode_label'] == 'SW Grism Time Series'


def test_get_configs_niriss_soss():
    configs = jwst._get_configs(instrument='niriss', obs_type='spectroscopy')
    assert len(configs) == 1
    inst = configs[0]
    assert inst['mode'] == 'soss'
    assert inst['mode_label'] == 'SOSS'


def test_get_configs_nirspec_bots():
    configs = jwst._get_configs(instrument='nirspec', obs_type='spectroscopy')
    assert len(configs) == 1
    inst = configs[0]
    assert inst['mode'] == 'bots'
    assert inst['mode_label'] == 'Bright Object Time Series'


def test_get_configs_acquisition():
    for instrument in jwst.get_instruments():
        configs = jwst._get_configs(instrument=instrument, obs_type='acquisition')
        assert len(configs) == 1
        inst = configs[0]
        assert inst['mode'] == 'target_acq'


def test_load_flux_rate_splines_all():
    # Extract all splines:
    flux_rate_splines, full_wells = jwst._load_flux_rate_splines()
    # Evaluate for one config:
    obs_label = 'lw_tsgrism_f444w_phoenix_k5v'
    spline = flux_rate_splines[obs_label]
    flux_rate = spline(8.351)
    full_well = full_wells[obs_label]

    expected_full_well = 58100.001867429375
    expected_flux_rate = 3.114047796362145
    np.testing.assert_allclose(flux_rate, expected_flux_rate)
    np.testing.assert_allclose(full_well, expected_full_well)


def test_load_flux_rate_splines_single():
    obs_label = 'lw_tsgrism_f444w_phoenix_k5v'
    flux_rate_spline, full_well = jwst._load_flux_rate_splines(obs_label)
    flux_rate = flux_rate_spline(8.351)

    expected_full_well = 58100.001867429375
    expected_flux_rate = 3.114047796362145
    np.testing.assert_allclose(flux_rate, expected_flux_rate)
    np.testing.assert_allclose(full_well, expected_full_well)


def test_load_flux_rate_splines_not_found():
    obs_label = 'nope_lw_tsgrism_f444w_phoenix_k5v'
    flux_rate_spline, full_well = jwst._load_flux_rate_splines(obs_label)
    assert flux_rate_spline is None
    assert full_well is None




# Copyright (c) 2025 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

__all__ = [
    'fetch_trexolist',
    'update_exoplanet_archive',
    'fetch_nasa_confirmed_targets',
    'fetch_gaia_targets',
    'fetch_nea_aliases',
    '_load_jwst_names',
]


import concurrent.futures
import multiprocessing as mp
from datetime import datetime, timezone
import os
import pickle
import re
import socket
import ssl
import urllib
import warnings


from astropy.coordinates import SkyCoord
import numpy as np
from astroquery.simbad import Simbad
from astroquery.vizier import Vizier
from astropy.table import Table
from astropy.units import arcsec, deg
from bs4 import BeautifulSoup
import pyratbay.constants as pc
import requests

from ..utils import ROOT
from .catalogs import (
    load_targets, load_trexolists, load_programs, load_aliases,
    _group_by_target,
)
from . import utils as u
from . import target as tar
from .target import Target


def format_nea_entry(entry):
    """
    Have TOI entries the same keys as PS entries.
    Turn None values into np.nan.
    Calculate stellar mass from logg-rstar for TESS candidates
    """
    if 'toi' in entry.keys():
        entry['hostname'] = f"TOI-{entry['toipfx']}"
        entry['st_met'] = np.nan
        entry['sy_kmag'] = np.nan
        logg = entry['st_logg'] if entry['st_logg'] is not None else np.nan
        rstar = entry['st_rad'] if entry['st_rad'] is not None else np.nan
        entry['st_mass'] = 10**logg * (rstar*pc.rsun)**2 / pc.G / pc.msun

        entry['pl_name'] = f"TOI-{entry['toi']}"
        entry['pl_masse'] = np.nan
        entry['pl_orbsmax'] = np.nan
        entry['pl_ratdor'] = np.nan
        entry['pl_ratror'] = np.sqrt(entry.pop('pl_trandep')*pc.ppm)
        entry['pl_trandur'] = entry.pop('pl_trandurh')
        entry['pl_msinie'] = np.nan
    else:
        entry['pl_eqt'] = np.nan

    # Patch
    if entry['st_mass'] == 0.0:
        entry['st_mass'] = np.nan

    # Replace None with np.nan
    for key in entry.keys():
        if entry[key] is None:
            entry[key] = np.nan
    return entry


def get_children(host_aliases, planet_aliases):
    """
    Cross check a dictionary of star and planet aliases to see
    whether the star is the host of the planets.
    """
    # get all planet aliases minus the 'letter' identifier
    planet_aka = u.invert_aliases(planet_aliases)
    for planet, aliases in planet_aka.items():
        aka = []
        for alias in aliases:
            len_letter = len(u.get_letter(alias))
            aka.append(alias[0:-len_letter])
        planet_aka[planet] = aka

    # cross_check with host aliases
    children = []
    for planet, aliases in planet_aka.items():
        if np.any(np.isin(aliases, host_aliases)):
            children.append(planet)

    aliases = {
        alias:planet
        for alias,planet in planet_aliases.items()
        if planet in children
    }
    return aliases


def save_catalog(targets, catalog_file):
    """
    Write data from a catalog of targets to a plain-text file.
    Targets will be sorted by host name and then by planet name.
    """
    # Save as plain text:
    with open(catalog_file, 'w') as f:
        f.write(
            '# > host: RA(deg) dec(deg) Ks_mag '
            'rstar(rsun) mstar(msun) teff(K) log_g metallicity(dex)\n'
            '# planet: T14(h) rplanet(rearth) mplanet(mearth) '
            'semi-major_axis(AU) period(d) t_eq(K) is_min_mass\n'
        )
        hosts = [target.host for target in targets]
        planets = [target.planet for target in targets]
        isort = np.lexsort((planets, hosts))
        host = ''
        for idx in isort:
            target = targets[idx]
            planet = target.planet
            ra = f'{target.ra:.7f}'
            dec = f'{target.dec:.7f}'
            ks_mag = f'{target.ks_mag:.3f}'
            teff = f'{target.teff:.1f}'
            rstar = f'{target.rstar:.3f}'
            mstar = f'{target.mstar:.3f}'
            logg = f'{target.logg_star:.2f}'
            metal = f'{target.metal_star:.2f}'
            rplanet = f'{target.rplanet:.3f}'
            mplanet = f'{target.mplanet:.3f}'
            transit_dur = f'{target.transit_dur:.3f}'
            sma = f'{target.sma:.4f}'
            period = f'{target.period:.5f}'
            teq = f'{target.eq_temp:.1f}'
            is_min_mass = int(target.is_min_mass)
            if target.host != host:
                host = target.host
                f.write(
                    f">{host}: {ra} {dec} {ks_mag} "
                    f"{rstar} {mstar} {teff} {logg} {metal}\n",
                )
            f.write(
                f" {planet}: {transit_dur} {rplanet} {mplanet} "
                f"{sma} {period} {teq} {is_min_mass}\n",
            )


def update_exoplanet_archive(from_scratch=False):
    """
    Fetch confirmed and TESS-candidate targets from the NASA
    Exoplanet archive.

    Parameters
    ----------
    from_scratch: Bool
        If True, fetch all aliases from scratch.  The run will
        take much longer but it will populate the local .pickle files
        with all known aliases, making later runs more efficient.

    Examples
    --------
    >>> import gen_tso.catalogs as cat
    >>> cat.update_exoplanet_archive()
    """
    # Update trexolist database
    fetch_trexolist()

    # NEA confirmed targets
    print('\nFetching confirmed planets from the NASA archive')
    new_targets = fetch_nasa_confirmed_targets()
    if from_scratch:
        new_targets = None
    print('Fetching confirmed planets aliases')
    fetch_confirmed_aliases(new_targets)

    # NEA TESS candidate targets
    print('Fetching TESS candidate planets from the NASA archive')
    new_targets = fetch_nasa_tess_candidates()
    if from_scratch:
        new_targets = None
    print('Fetching TESS candidates aliases')
    fetch_tess_aliases(new_targets)
    crosscheck_tess_candidates()

    today = datetime.now(timezone.utc)
    with open(f'{ROOT}/data/last_updated_nea.txt', 'w') as f:
        f.write(f'{today.year}_{today.month:02}_{today.day:02}')

    # Update aliases list
    curate_aliases()
    print('Exoplanet data is up to date')


def _load_jwst_names(grouped=False):
    """
    Keep track of JWST target aliases from programs to cross-check
    from gen_tso.catalogs import load_trexolists, load_programs
    from gen_tso.catalogs import *
    """
    trexo_data = load_trexolists(grouped=False)
    observations = load_programs(grouped=False)
    if not grouped:
        jwst_names = np.unique([obs['target'] for obs in trexo_data])
        known_targets = np.unique([obs['target'] for obs in observations])
        jwst_names = np.union1d(jwst_names, known_targets).tolist()
        return jwst_names

    # grouped by target
    all_obs = trexo_data + observations
    observations = _group_by_target(all_obs)
    jwst_aliases = []
    for obs_group in observations:
        names = np.unique([obs['target'] for obs in obs_group]).tolist()
        jwst_aliases.append(names)

    return jwst_aliases


def curate_aliases():
    """
    Thin down all_aliases.pickle file to the essentials.
    Save as .txt, which is shipped with gen_tso.
    """
    with open(f'{ROOT}data/nea_aliases.pickle', 'rb') as handle:
        aliases = pickle.load(handle)
    with open(f'{ROOT}data/tess_aliases.pickle', 'rb') as handle:
        tess_aliases = pickle.load(handle)
    aliases.update(tess_aliases)

    # Ensure to match against NEA host names for jwst targets
    jwst_names = _load_jwst_names()
    for host,system in aliases.items():
        is_in = np.isin(system['host_aliases'], jwst_names)
        if np.any(is_in) and system['host'] not in jwst_names:
            jwst_names.append(system['host'])

    prefixes = jwst_names
    prefixes += ['WASP', 'KELT', 'HAT', 'MASCARA', 'TOI', 'XO', 'TrES']
    keep_aliases = {}
    for host,system in aliases.items():
        for alias,planet in system['planet_aliases'].items():
            alias_host = u.get_host(alias)
            for prefix in prefixes:
                if alias.startswith(prefix) and alias != planet:
                    keep_aliases[alias] = planet
            if alias not in keep_aliases and alias_host == host and alias != planet:
                keep_aliases[alias] = planet

    aka = u.invert_aliases(keep_aliases)
    to_remove = []
    for name, aliases in aka.items():
        # Keep lettered aliases
        # Keep candidate aliases if lettered alias does not exist
        lettered = [
            u.get_host(alias)
            for alias in aliases
            if u.is_letter(alias)
        ]
        if u.is_letter(name):
            lettered.append(u.get_host(name))
        aliases = [
            alias
            for alias in aliases
            if u.is_letter(alias) or (u.get_host(alias) not in lettered)
        ]
        aka[name] = aliases

        if len(aliases) == 0:
            to_remove.append(name)

    for name in to_remove:
        aka.pop(name)

    sorted_names = sorted(list(aka))
    with open(f'{ROOT}data/target_aliases.txt', 'w') as f:
        for name in sorted_names:
            # TBD: catch this, programatically
            if name == 'LP 261-75 C':
                continue
            aliases = sorted(aka[name])
            str_aliases = ','.join(aliases)
            f.write(f'{name}:{str_aliases}\n')


def fetch_trexolist():
    """
    Examples
    --------
    >>> import gen_tso.catalogs as cat
    >>> cat.fetch_trexolist()
    """
    # Fetch the data:
    url = "https://www.stsci.edu/~nnikolov/TrExoLiSTS/JWST/03_trexolists_extended.csv"
    query_parameters = {}
    response = requests.get(url, params=query_parameters)

    if not response.ok:
        raise ValueError('Could not download TrExoLiSTS database')

    trexolists_path = f'{ROOT}data/trexolists.csv'
    with open(trexolists_path, mode="wb") as file:
        file.write(response.content)

    # Fetch the last-update date:
    url = "https://www.stsci.edu/~nnikolov/TrExoLiSTS/JWST/trexolists.html"
    response = requests.get(url)
    soup = BeautifulSoup(response.text, 'html.parser')
    last_update_tag = soup.find('h3', string=lambda text: 'Last update' in text)
    last_update_text = last_update_tag.get_text(strip=True)
    date = datetime.strptime(last_update_text[13:], '%Y-%m-%d %H:%M:%S')
    with open(f'{ROOT}data/last_updated_trexolist.txt', 'w') as f:
        f.write(f'{date.year}_{date.month:02}_{date.day:02}')


def fetch_nasa_confirmed_targets():
    """
    Fetch (HTTP web request) the entire NASA Exoplanet Archive database
    for confirmed planets (there is another one for TESS candidates)

    Returns
    -------
    new_targets: List of strings
        List of target names flagged as updated by the NEA
        since the last update (rowupdate column).

    See also
    --------
    - fetch_nasa_tess_candidates()  to fetch the TESS candidates database

    Examples
    --------
    >>> import gen_tso.catalogs as cat
    >>> new_targets = cat.fetch_nasa_confirmed_targets()
    """
    # Fetch all planetary system entries
    r = requests.get(
        "https://exoplanetarchive.ipac.caltech.edu/TAP/sync?query="
        "select+hostname,pl_name,default_flag,rowupdate,sy_kmag,sy_pnum,"
        "ra,dec,st_teff,st_logg,st_met,st_rad,st_mass,st_age,pl_trandur,"
        "pl_orbper,pl_orbsmax,pl_rade,pl_masse,pl_msinie,pl_ratdor,pl_ratror+"
        "from+ps+"
        "&format=json"
    )
    if not r.ok:
        raise ValueError("Something's not OK")
    resp = [format_nea_entry(entry) for entry in r.json()]

    hosts = np.unique([entry['hostname'] for entry in resp])
    default_flags = np.array([entry['default_flag'] for entry in resp])
    planet_entries = np.array([entry['pl_name'] for entry in resp])
    planet_names, idx, counts = np.unique(
        planet_entries,
        return_index=True,
        return_counts=True,
    )

    targets = []
    # Make list of unique entries
    # Group by host such that planets share same host-star values
    for host in hosts:
        children, counts = np.unique(
            [entry['pl_name'] for entry in resp if entry['hostname']==host],
            return_counts=True,
        )
        planets = []
        n_dups = []
        for name in children:
            idx_entry = np.where(planet_entries==name)[0]
            entries = [Target(resp[i]) for i in idx_entry]
            j = np.where(default_flags[idx_entry])[0][0]
            target = entries.pop(j)
            tar.rank_planets(target, entries)
            planets.append(target)
            n_dups.append(len(idx_entry))
        # Solve stellar parameters (all planets must have the 'same' host)
        star = tar.solve_host(planets, n_dups)

        # Now, re-do each planet, but using the single host properties
        for name in children:
            idx_entry = np.where(planet_entries==name)[0]
            entries = [Target(resp[i]) for i in idx_entry]
            # Update with star props
            for i in range(len(entries)):
                entries[i].copy_star(star)
            j = np.where(default_flags[idx_entry])[0][0]
            target = entries.pop(j)
            tar.rank_planets(target, entries)
            target._update_dates = [resp[i]['rowupdate'] for i in idx_entry]
            targets.append(target)

    # Find new and updated targets:
    catalog_file = f'{ROOT}data/nea_data.txt'
    if os.path.exists(catalog_file):
        current_targets = [
            target.planet for target in load_targets('nea_data.txt')
        ]
    else:
        current_targets = []

    update_file = f'{ROOT}/data/last_updated_nea.txt'
    if os.path.exists(update_file):
        with open(update_file, 'r') as f:
            date = f.readline().strip()
            last_nasa = datetime.strptime(date,'%Y_%m_%d')
    else:
        last_nasa = datetime.strptime('1990', '%Y')

    new_targets = []
    n_new, n_updated = 0, 0
    for target in targets:
        if target.planet not in current_targets:
            new_targets.append(target.planet)
            n_new += 1
        else:
            is_new = [
                datetime.strptime(date, '%Y-%m-%d') > last_nasa
                for date in target._update_dates
                if isinstance(date, str)
            ]
            if np.any(is_new):
                n_updated += 1
                new_targets.append(target.planet)
    print(
        f'There are {n_new} new and {n_updated} updated confirmed '
        f'targets since {last_nasa.strftime("%Y-%m-%d")}'
    )

    # Save outputs
    save_catalog(targets, catalog_file)
    return new_targets


def fetch_nasa_tess_candidates():
    """
    Fetch entries in the NEA TESS candidates table.
    Remove already confirmed targets.

    Returns
    -------
    new_targets: List of strings
        Targets flagged by NEA to be updated since the last fetch.

    Examples
    --------
    >>> from gen_tso.catalogs import fetch_catalogs as fetch_cat
    >>> new_targets = fetch_cat.fetch_nasa_tess_candidates()
    >>> fetch_cat.fetch_tess_aliases(new_targets)
    >>> fetch_cat.crosscheck_tess_candidates()
    """
    r = requests.get(
        "https://exoplanetarchive.ipac.caltech.edu/TAP/sync?query="
        "select+toi,toipfx,pl_trandurh,pl_trandep,pl_rade,pl_eqt,ra,dec,"
        "st_tmag,st_teff,st_logg,st_rad,pl_orbper,tfopwg_disp,rowupdate+"
        "from+toi+"
        "&format=json"
    )
    if not r.ok:
        raise ValueError("Something's not OK")

    entries = [format_nea_entry(entry) for entry in r.json()]
    ntess = len(entries)
    status = [entry['tfopwg_disp'] for entry in entries]
    dates = [entry['rowupdate'][0:10] for entry in entries]

    # Discard confirmed planets:
    targets = load_targets('nea_data.txt')
    confirmed_targets = [target.planet for target in targets]
    confirmed_hosts = [target.host for target in targets]

    # Use multiplicity to vet known targets:
    hosts = [entry['toipfx'] for entry in entries]
    u_hosts, counts = np.unique(hosts, return_counts=True)
    u_hosts = list(u_hosts)
    multiplicity = [
        counts[u_hosts.index(entry['toipfx'])] for entry in entries
    ]
    u_hosts, counts = np.unique(confirmed_hosts, return_counts=True)
    u_hosts = list(u_hosts)
    known_multiplicity = [
        counts[u_hosts.index(target.host)] for target in targets
    ]

    # Get aliases of confirmed planets:
    planet_aliases = {}
    host_aliases = {}

    aliases_file = f'{ROOT}data/target_aliases.txt'
    if os.path.exists(aliases_file):
        known_aliases = load_aliases('system')
        for host, system in known_aliases.items():
            if host in confirmed_hosts:
                planet_aliases.update(system['planet_aliases'])
                for alias in system['host_aliases']:
                    host_aliases[alias] = host

    aliases_file = f'{ROOT}data/nea_aliases.pickle'
    if os.path.exists(aliases_file):
        with open(aliases_file, 'rb') as handle:
            known_aliases = pickle.load(handle)
        for host, system in known_aliases.items():
            planet_aliases.update(system['planet_aliases'])
            for alias in system['host_aliases']:
                host_aliases[alias] = host

    j, k, l = 0, 0, 0
    tess_targets = []
    last_updated = []
    for i in range(ntess):
        target = Target(entries[i])
        # Update names if possible:
        if target.host in host_aliases:
            target.host = host_aliases[target.host]
        if target.planet in planet_aliases:
            target.planet = planet_aliases[target.planet]

        if target.planet in confirmed_targets:
            j += 1
            continue
        if target.host in confirmed_hosts:
            idx = confirmed_hosts.index(target.host)
            if multiplicity[i] == known_multiplicity[idx]:
                continue
            # Update with star props
            k += 1
            target.copy_star(targets[idx])
        if status[i] in ['FA', 'FP']:
            l += 1
            continue
        tess_targets.append(target)
        last_updated.append(datetime.strptime(dates[i], '%Y-%m-%d'))

    # Save temporary data (still need to hunt for Ks mags):
    catalog_file = f'{ROOT}data/tess_candidates_tmp.txt'
    save_catalog(tess_targets, catalog_file)

    with open(f'{ROOT}/data/last_updated_nea.txt', 'r') as f:
        last_nasa = datetime.strptime(f.readline().strip(),'%Y_%m_%d')
    new_targets = [
        target.planet
        for target,last in zip(tess_targets,last_updated)
        if last > last_nasa or target.host in confirmed_hosts
    ]
    return new_targets


def fetch_nea_aliases(targets):
    """
    Fetch target aliases as known by https://exoplanetarchive.ipac.caltech.edu/

    Note 1: a search of a planet or stellar target returns the
        aliases for all bodies in that planetary system.
    Note 2: there might be more than one star per system

    Parameters
    ----------
    targets: String or 1D iterable of strings
        Target(s) to fetch from the NEA database.

    Returns
    -------
    host_aliases_list: 1D list of dictionaries
        List of host-star aliases for each target.
    planet_aliases_list: 1D list of dictionaries
        List of planetary aliases for each target.

    Examples
    --------
    >>> import gen_tso.catalogs as cat
    >>> targets = ['WASP-8 b', 'KELT-7', 'HD 189733']
    >>> host_aliases, planet_aliases = cat.fetch_nea_aliases(targets)

    >>> host_aliases, planet_aliases = cat.fetch_nea_aliases('WASP-00')
    """
    if isinstance(targets, str):
        targets = [targets]
    ntargets = len(targets)

    urls = np.array([
        'https://exoplanetarchive.ipac.caltech.edu/cgi-bin/Lookup/'
        f'nph-aliaslookup.py?objname={urllib.parse.quote(target)}'
        for target in targets
    ])

    def fetch_url(url):
        try:
            response = requests.get(url)
            return response
        except:
            return None

    fetch_status = np.tile(2, ntargets)
    responses = np.tile({}, ntargets)
    n_attempts = 0
    while np.any(fetch_status>0) and n_attempts < 10:
        n_attempts += 1
        mask = fetch_status > 0
        with concurrent.futures.ThreadPoolExecutor() as executor:
            results = list(executor.map(fetch_url, urls[mask]))

        j = 0
        for i in range(ntargets):
            if fetch_status[i] <= 0:
                continue
            r = results[j]
            j += 1
            if r is None:
                continue
            if not r.ok:
                warnings.warn(f"Alias fetching failed for '{targets[i]}'")
                fetch_status[i] -= 1
                continue
            responses[i] = r.json()
            fetch_status[i] = 0
        fetched = np.sum(fetch_status <= 0)
        print(f'Fetched {fetched}/{ntargets} entries on try {n_attempts}')

    host_aliases_list = []
    planet_aliases_list = []
    for i,resp in enumerate(responses):
        if resp == {}:
            print(f"NEA alias fetching failed for '{targets[i]}'")
            host_aliases_list.append({})
            planet_aliases_list.append({})
            continue
        if resp['manifest']['lookup_status'] == 'System Not Found':
            print(f"NEA alias not found for '{targets[i]}'")
            host_aliases_list.append({})
            planet_aliases_list.append({})
            continue

        host_aliases = {}
        star_set = resp['system']['objects']['stellar_set']['stars']
        for star in star_set.keys():
            if 'is_host' not in star_set[star]:
                continue
            for alias in star_set[star]['alias_set']['aliases']:
                host_aliases[alias] = star
        host_aliases_list.append(host_aliases)

        planet_aliases = {}
        planet_set = resp['system']['objects']['planet_set']['planets']
        for planet in planet_set.keys():
            for alias in planet_set[planet]['alias_set']['aliases']:
                planet_aliases[alias] = planet
        planet_aliases_list.append(planet_aliases)

    return host_aliases_list, planet_aliases_list


def fetch_simbad_aliases(targets):
    """
    Fetch target aliases and Ks-band magnitude from the Simbad database.

    Parameters
    ----------
    targets: string or iterable of strings
        Name or list of names of astronomical targets to query in SIMBAD.

    Returns
    -------
    aliases: list of lists of str
        A list containing the Simbad aliases (list) for each target.
        If a target is not found, the corresponding list is empty.
    ks_mag: 1D float array
        Ks-band magnitude for each targets.
        If a target is not found, the corresponding magnitude is a np.nan.

    Examples
    --------
    >>> from gen_tso.catalogs.update_catalogs import fetch_simbad_aliases
    >>> # Single-planet search
    >>> aliases, ks_mag = fetch_simbad_aliases('WASP-80')
    >>> print(aliases[0], ks_mag[0], sep='\n')
    ['Gaia DR3 4223507222112425344', 'StKM 2-1435', ..., 'NAME Petra']
    8.35099983215332

    >>> # Batch search
    >>> targets = ['Kepler-138', 'TOI-270', 'WASP-39']
    >>> aliases, ks_mag = fetch_simbad_aliases(targets)
    >>> print(ks_mag)
    [ 9.50599957  8.2510004  10.20199966]

    >>> # Not-found/non-existing targets returns
    >>> aliases, ks_mag = fetch_simbad_aliases('WASP-0')
    >>> print(aliases[0], ks_mag[0], sep='\n')
    []
    nan
    """
    if isinstance(targets, str):
        targets = [targets]
    Simbad.reset_votable_fields()
    Simbad.add_votable_fields("otype", "ids", 'K')

    query = Simbad.query_objects(targets)
    kmag = np.array(query['K'])

    host_aliases = []
    for host in query:
        aliases = host['ids'].split('|')
        # Object not found:
        if host['otype'] == '':
            host_aliases.append([])
            continue
        # Symbad thinks that some hosts are planets:
        if host['otype'] == 'Pl':
            aliases = []
            for alias in host['ids'].split('|'):
                if u.is_candidate(alias) or u.is_letter(alias):
                    alias = u.get_host(alias)
                elif alias[-1] == 'b':
                    alias = alias[:-1]
                aliases.append(alias)
        host_aliases.append(aliases)

    return host_aliases, kmag


def fetch_vizier_ks(target, verbose=True):
    """
    Query for a target in the 2MASS catalog via Vizier.

    Returns
    -------
    Ks_mag: Float
        The target's Ks magnitude.
        Return None if the target was not found in the catalog or
        could not be uniquely identified.

    Examples
    --------
    >>> fetch_vizier_ks('TOI-6927')
    >>> fetch_vizier_ks('2MASS J08024565+2139348')
    >>> fetch_vizier_ks('Gaia DR2 671023360793596672')
    """
    # 2mass catalog
    catalog = 'II/246/out'
    vizier = Vizier(
        catalog=catalog,
        columns=['RAJ2000', 'DEJ2000', '2MASS', 'Kmag'],
        keywords=['Stars'],
    )

    result = vizier.query_object(target, radius=0.5*arcsec)
    n_entries = np.size(result)
    if n_entries == 0:
        print(f"Target not found: '{target}'")
        return np.nan

    data = result[catalog].as_array().data
    if n_entries == 1:
        return data[0][3]
    elif n_entries > 1 and target.startswith('2MASS'):
        # find by name
       for row in data:
           if row[2] == target[-16:]:
               return row[3]
    elif n_entries > 1:
        print(f"Target could not be uniquely identified: '{target}'")
        return np.nan
    return np.nan


def fetch_aliases(hosts, output_file=None, known_aliases=None):
    """
    Fetch aliases from the NEA and Simbad databases for a list
    of host stars.  Store output dictionary of aliases to pickle file.

    Parameters
    ----------
    hosts: List of strings
        Host star names of targets to search.
    output_file: String
        If not None, save outputs to file (as pickle binary format).
    known_aliases: Dictionary
        Dictionary of known aliases, the new aliases will be added
        on top o this dictionary.

    Returns
    -------
    aliases: Dictionary
        Dictionary of aliases with one entry per system where
        the key is the host name.  Each entry is a dictionary
        containing:
            'host' the host name (string)
            'planets': array of planets in the system
            'host_aliases': list of host aliases
            'planet_aliases': dictionary of planets' aliases with
                (key,value) as (alias_name, name)

    Examples
    --------
    >>> import gen_tso.catalogs as cat
    >>> from gen_tso.utils import ROOT

    >>> # Confirmed targets
    >>> targets = cat.load_targets()
    >>> hosts = np.unique([target.host for target in targets])
    >>> output_file = f'{ROOT}data/nea_aliases.pickle'
    >>> aliases = cat.fetch_aliases(hosts, output_file)
    """
    if known_aliases is None:
        known_aliases = {}

    host_aliases, planet_aliases = fetch_nea_aliases(hosts)

    # Ensure to match against JWST program target names
    jwst_aliases = _load_jwst_names(grouped=True)
    jwst_names = np.concatenate(jwst_aliases)

    # Ensure I have the NEA-named host name:
    nhosts = len(hosts)
    host_names = np.array(hosts).tolist()
    for i in range(nhosts):
        if len(host_aliases[i]) == 0:
            continue
        hosts_aka = u.invert_aliases(host_aliases[i])
        for nea_host, h_aliases in hosts_aka.items():
            if hosts[i] in h_aliases:
                host_names[i] = nea_host
                break

    # Get aliases from Simbad:
    simbad_aliases, simbad_kmags = fetch_simbad_aliases(host_names)

    aliases = {}
    nhosts = len(hosts)
    for i in range(nhosts):
        host_name = host_names[i]
        if len(host_aliases[i]) == 0:
            continue
        # Isolate host-planet(s) aliases
        h_aliases = [
            alias
            for alias,host in host_aliases[i].items()
            if host == host_names[i]
        ]
        stars = np.unique(list(host_aliases[i].values()))
        if len(stars) == 1:
            p_aliases = planet_aliases[i].copy()
        else:
            p_aliases = get_children(h_aliases, planet_aliases[i])
        p_aliases = {
            re.sub(r'\s+', ' ', key): val
            for key,val in p_aliases.items()
            if u.is_letter(key) or u.is_candidate(key)
        }
        children_names = np.unique(list(p_aliases.values()))

        in_jwst = np.isin(h_aliases, jwst_names)
        if np.any(in_jwst):
            j_alias = np.array(h_aliases)[in_jwst]

        # Complement with Simbad aliases:
        new_aliases = []
        for alias in simbad_aliases[i]:
            alias = re.sub(r'\s+', ' ', alias)
            is_new = (
                alias in jwst_names or
                alias.startswith('G ') or
                alias.startswith('GJ ') or
                alias.startswith('Wolf ') or
                alias.startswith('2MASS ')
            )
            if is_new and alias not in h_aliases:
                new_aliases.append(alias)
                h_aliases.append(alias)

        # Add JWST host aliases:
        in_jwst = np.isin(h_aliases, jwst_names)
        if np.any(in_jwst):
            j_alias = np.array(h_aliases)[in_jwst][0]
            for j_aliases in jwst_aliases:
                if j_alias not in j_aliases:
                    continue
                h_aliases += [
                    alias
                    for alias in j_aliases
                    if alias not in h_aliases
                ]

        # Replicate host aliases as planet aliases:
        planet_aka = u.invert_aliases(p_aliases)
        for planet, pals in planet_aka.items():
            for host in h_aliases:
                letter = u.get_letter(planet)
                planet_name = f'{host}{letter}'
                # The conditions to add a target:
                is_new = planet_name not in pals
                # There is a planet or a candidate in list
                planet_exists = np.any([
                    u.get_host(p) == host and u.is_letter(p)
                    for p in pals
                ])
                candidate_exists = np.any([
                    u.get_host(p) == host and u.is_candidate(p)
                    for p in pals
                ])
                # Do not downgrade planet -> candidate
                not_downgrade = not (
                    u.is_candidate(planet_name) and
                    planet_exists
                )
                # No previous alias (hold-off TESS names)
                new_entry = (
                    not planet_exists and
                    not candidate_exists and
                    not planet_name.startswith('TOI')
                )
                # There is a letter version of it with same root
                letter_exists = np.any([
                    p.startswith(host) and u.is_letter(p)
                    for p in pals
                ])
                # Upgrade candidate->planet only if is lettered anywhere else
                upgrade = (
                    u.is_letter(planet_name) and
                    candidate_exists and
                    letter_exists
                )
                if is_new and not_downgrade and (new_entry or upgrade):
                    p_aliases[planet_name] = planet

        system = {
            'host': host_name,
            'planets': children_names,
            'host_aliases': h_aliases,
            'planet_aliases': p_aliases,
        }
        aliases[host_name] = system


    # Add previously known aliases (but give priority to the new ones)
    for host in list(known_aliases):
        if host not in aliases:
            aliases[host] = known_aliases[host]

    if output_file is not None:
        with open(output_file, 'wb') as handle:
            pickle.dump(aliases, handle, protocol=4)

    return aliases


def fetch_confirmed_aliases(new_targets=None):
    """
    Fetch aliases for NEA confirmed planets.
    Save results to a binary file 'nea_aliases.pickle'

    Parameters
    ----------
    new_targets: List of strings
        If not None, only fetch aliases for the given targets.


    Examples
    --------
    >>> from gen_tso.catalogs import fetch_catalogs as fetch_cat
    >>> new_targets = fetch_cat.fetch_nasa_confirmed_targets()
    >>> aliases = fetch_cat.fetch_confirmed_aliases(new_targets)
    """
    known_targets = load_targets()
    # Search aliases by host star
    if new_targets is None:
        hosts = np.unique([target.host for target in known_targets])
    else:
        hosts = np.unique([
            target.host for target in known_targets
            if target.planet in new_targets
        ])

    # Get previously known aliases
    known_aliases = {}
    aliases_file = f'{ROOT}data/target_aliases.txt'
    if os.path.exists(aliases_file):
        known_aliases = load_aliases('system')
    output_file = f'{ROOT}data/nea_aliases.pickle'
    if os.path.exists(output_file):
        with open(output_file, 'rb') as handle:
            prev_aliases = pickle.load(handle)
        for host,system in prev_aliases.items():
            known_aliases[host] = system

    known_hosts = np.unique([target.host for target in known_targets])
    for host in list(known_aliases):
        if host not in known_hosts:
            known_aliases.pop(host)

    # Get new aliases
    aliases = fetch_aliases(hosts, output_file, known_aliases)
    return aliases


def fetch_tess_aliases(new_targets=None):
    """
    Get TESS candidate aliases.
    You want to run fetch_nasa_tess_candidates() before this one.
    And then follow up with crosscheck_tess_candidates()

    Parameters
    ----------
    new_targets: List of strings
        If not None, only fetch aliases for the given targets.

    Examples
    --------
    >>> from gen_tso.catalogs import fetch_catalogs as fetch_cat
    >>> new_targets = fetch_cat.fetch_nasa_tess_candidates()
    >>> aliases = fetch_cat.fetch_tess_aliases(new_targets)
    >>> fetch_cat.crosscheck_tess_candidates()
    """
    candidates = load_targets('tess_candidates_tmp.txt')
    if new_targets is None:
        new_targets = np.unique([target.planet for target in candidates])

    if os.path.exists(f'{ROOT}data/tess_data.txt'):
        known_candidates = load_targets('tess_data.txt')
        known_tess = [target.planet for target in known_candidates]
    else:
        known_tess = []

    # New TESS hosts that are not in confirmed list
    hosts = np.unique([
        target.host
        for target in candidates
        if target.planet in new_targets or target.planet not in known_tess
    ])


    # Get previously known aliases
    known_aliases = {}
    aliases_file = f'{ROOT}data/target_aliases.txt'
    if os.path.exists(aliases_file):
        known_aliases = load_aliases('system')
    output_file = f'{ROOT}data/tess_aliases.pickle'
    # Get previously known aliases
    if os.path.exists(output_file):
        with open(output_file, 'rb') as handle:
            prev_aliases = pickle.load(handle)
        for host,system in prev_aliases.items():
            known_aliases[host] = system

    known_hosts = np.unique([target.host for target in candidates])
    for host in list(known_aliases):
        if host not in known_hosts:
            known_aliases.pop(host)

    # Get new aliases
    aliases = fetch_aliases(hosts, output_file, known_aliases)
    return aliases


def crosscheck_tess_candidates(ncpu=None):
    """
    Do a final TESS-confirmed cross-check and hunt for their Ks mag.
    Write output to tess_data.txt file.

    Before calling this function, you want to run
    fetch_nasa_tess_candidates() and crosscheck_tess_candidates()

    Examples
    --------
    >>> from gen_tso.catalogs import fetch_catalogs as fetch_cat
    >>> new_targets = fetch_cat.fetch_nasa_tess_candidates()
    >>> fetch_cat.fetch_tess_aliases(new_targets)
    >>> fetch_cat.crosscheck_tess_candidates()
    """
    if ncpu is None:
        ncpu = mp.cpu_count()

    planet_aliases = {}
    host_aliases = {}
    # previously known aliases
    aliases_file = f'{ROOT}data/target_aliases.txt'
    if os.path.exists(aliases_file):
        known_aliases = load_aliases('system')
        for host, system in known_aliases.items():
            planet_aliases.update(system['planet_aliases'])
            for alias in system['host_aliases']:
                host_aliases[alias] = host
    # known confirmed aliases
    with open(f'{ROOT}data/nea_aliases.pickle', 'rb') as handle:
        known_aliases = pickle.load(handle)
    for host, system in known_aliases.items():
        planet_aliases.update(system['planet_aliases'])
        for alias in system['host_aliases']:
            host_aliases[alias] = host
    # known TESS aliases
    with open(f'{ROOT}data/tess_aliases.pickle', 'rb') as handle:
        tess_aliases = pickle.load(handle)
    for host, system in tess_aliases.items():
        planet_aliases.update(system['planet_aliases'])
        for alias in system['host_aliases']:
            host_aliases[alias] = host

    # Identity alias for targets without aliases
    candidates = load_targets('tess_candidates_tmp.txt')
    for target in candidates:
        if target.host not in host_aliases:
            host_aliases[target.host] = target.host
        if target.planet not in planet_aliases:
            planet_aliases[target.planet] = target.planet
    aka = u.invert_aliases(host_aliases)

    # Cross-check with confirmed targets:
    confirmed_planets = [target.planet for target in load_targets()]
    candidates = [
        target for target in candidates
        if planet_aliases[target.planet] not in confirmed_planets
    ]
    # Make sure names are NEA names:
    for i,target in enumerate(candidates):
        target.host = host_aliases[target.host]
        target.planet = planet_aliases[target.planet]


    # Now I need to collect the Ks-band magnitudes:
    # Zeroth idea, check if I already have the Ks mag:
    if os.path.exists(f'{ROOT}data/tess_data.txt'):
        known_candidates = load_targets('tess_data.txt')
        known_hosts = [target.host for target in known_candidates]
        for target in candidates:
            if target.host in known_hosts:
                idx = known_hosts.index(target.host)
                target.ks_mag = known_candidates[idx].ks_mag

    # First idea, search in simbad using best known alias to get Ks magnitude
    catalogs = ['2MASS', 'Gaia DR3', 'Gaia DR2', 'TOI']
    names = [
        u.select_alias(aka[target.host], catalogs)
        for i,target in enumerate(candidates)
    ]
    aliases, kmags = fetch_simbad_aliases(names)
    for i,target in enumerate(candidates):
        if not np.isfinite(target.ks_mag) and np.isfinite(kmags[i]):
            target.ks_mag = kmags[i]

    # Plan B, batch search in vizier/2MASS catalog:
    hosts = np.array([
        u.select_alias(aka[target.host], catalogs, target.host)
        for target in candidates
    ])
    missing_targets = [
        target for i,target in enumerate(candidates)
        if np.isnan(target.ks_mag)
        if hosts[i].startswith('2MASS')
    ]
    ra = np.array([target.ra for target in missing_targets]) * deg
    dec = np.array([target.dec for target in missing_targets]) * deg

    catalog = 'II/246/out'
    vizier = Vizier(
        catalog=catalog,
        columns=['RAJ2000', 'DEJ2000', '2MASS', 'Kmag'],
    )
    two_mass_targets = Table(
        [ra, dec],
        names=('_RAJ2000', '_DEJ2000'),
    )
    results = vizier.query_region(two_mass_targets, radius=5.0*arcsec)
    data = results[catalog].as_array().data
    vizier_names = [f'2MASS J{d[3]}' for d in data]
    vizier_ks_mag = [d[4] for d in data]

    for target in candidates:
        host = u.select_alias(aka[target.host], catalogs)
        if host in vizier_names:
            idx = vizier_names.index(host)
            target.ks_mag = vizier_ks_mag[idx]

    # Last resort, scrap from the NEA website
    missing_hosts = np.unique([
        target.host for target in candidates
        if np.isnan(target.ks_mag)
    ])
    with mp.get_context('fork').Pool(ncpu) as pool:
        scrap_ks = pool.map(scrap_nea_kmag, missing_hosts)
    for target in candidates:
        if target.host in missing_hosts:
            idx = list(missing_hosts).index(target.host)
            target.ks_mag = scrap_ks[idx]

    # Save as plain text:
    catalog_file = f'{ROOT}data/tess_data.txt'
    save_catalog(candidates, catalog_file)


def scrap_nea_kmag(target):
    """
    >>> target = 'TOI-5290'
    >>> target = 'TOI-4345'
    >>> scrap_nea_kmag(target)
    """
    response = requests.get(
        url=f'https://exoplanetarchive.ipac.caltech.edu/overview/{target}',
    )
    kmag = np.nan
    if not response.ok:
        print(f'ERROR {target}')
        return kmag

    soup = BeautifulSoup(response.content, 'html.parser')
    pm = '±'
    for dd in soup.find_all('dd'):
        texts = dd.get_text().split()
        if 'mKs' in texts:
            #print(target, dd.text.split())
            kmag_text = texts[-1]
            if kmag_text == '---':
                kmag = 0.0
            elif pm in kmag_text:
                kmag = float(kmag_text[0:kmag_text.find(pm)])
            else:
                kmag = float(kmag_text)
    return kmag


def fetch_gaia_targets(
        ra_source, dec_source, max_separation=80.0, raise_errors=True,
    ):
    """
    Search for Gaia DR3 stellar sources around a given target.

    Parameters
    ----------
    ra_source: Float
        Right ascension (deg) of target to query around.
    dec_source: Float
        Declination (deg) of target to query around.
    max_angular_distance: Float
        Maximum angular distance from target to consider (arcsec).
        Consider that the visit splitting distance for NIRSpec TA is 38"
    raise_errors: Bool
        If True and there was an error while requesting the data,
        raise the error.
        If False and there was an error, print error to screen and
        return a string identifying some known error types.

    Returns
    -------
    names: 1D string array
        Gaia DR3 stellar source names within max_separation from target
    G_mag: 1D float array
        Gaia G magnitude of found stellar sources
    teff: 1D float array
        Effective temperature (K) of found stellar source
    log_g: 1D float array
        log(g) of found stellar sources
    ra: 1D float array
        Right ascension (deg) of found stellar sources
    dec: 1D float array
        Declination (deg) of found stellar sources
    separation: 1D float array
        Angular separation (arcsec) of found stellar sources from target

    Examples
    --------
    >>> import gen_tso.catalogs as cat

    >>> # Stellar sources around WASP-69:
    >>> ra_source = 315.0259661
    >>> dec_source = -5.094857
    >>> cat.fetch_gaia_targets(ra_source, dec_source)
    """
    # Moved inside function to avoid hanging at import time
    # (when astroquery is not reachable)
    from astroquery.gaia import Gaia
    max_sep_degrees = max_separation / 3600.0

    try:
        job = Gaia.launch_job_async(
            f"""
            SELECT * \
            FROM gaiadr3.gaia_source \
            WHERE CONTAINS(\
                POINT(gaiadr3.gaia_source.ra, gaiadr3.gaia_source.dec),
                CIRCLE({ra_source}, {dec_source}, {max_sep_degrees}))=1;""",
            dump_to_file=False,
        )
    except Exception as e:
        err_text = f"Gaia astroquery request failed with {e.__class__.__name__}"
        if isinstance(e, socket.gaierror):
            print(
                f"\n{err_text}\n{str(e)}\n"
                "Likely there's no internet connection at the moment\n"
            )
            exception = 'gaierror'
        elif isinstance(e, requests.exceptions.HTTPError):
            print(
                f"\n{err_text}\n{str(e)}\n"
                f"Probably the ESA server is down at the moment\n"
            )
            exception = 'gaierror'
        elif isinstance(e, ssl.SSLError):
            print(
                f"\n{err_text}\n{str(e)}\n"
                "If you got a 'SSL: CERTIFICATE_VERIFY_FAILED' error on an "
                "OSX machine, try following the steps on this link: "
                "https://stackoverflow.com/a/42334357 which will point you to "
                "the ReadMe.rtf file in your Applications/Python 3.X folder\n"
            )
            exception = 'ssl'
        else:
            print(f"\n{err_text}\n{str(e)}")
            exception = 'other'
        if raise_errors:
            raise e
        return exception

    resp = job.get_results()
    targets = resp[~resp['teff_gspphot'].mask]

    c1 = SkyCoord(ra_source, dec_source, unit='deg', frame='icrs')
    separation = []
    for i,target in enumerate(targets):
        c2 = SkyCoord(target['ra'], target['dec'], unit='deg', frame='icrs')
        sep = c1.separation(c2).to('arcsec').value
        separation.append(sep)

    sep_isort = np.argsort(separation)
    for i,idx in enumerate(sep_isort):
        target = targets[idx]

    return (
            targets['designation'].data.data[sep_isort],
            targets['phot_g_mean_mag'].data.data[sep_isort],
            targets['teff_gspphot'].data.data[sep_isort],
            targets['logg_gspphot'].data.data[sep_isort],
            targets['ra'].data.data[sep_isort],
            targets['dec'].data.data[sep_isort],
            np.array(separation)[sep_isort],
    )




# Copyright (c) 2025 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

import pytest
import gen_tso.catalogs as cat
import numpy as np
from gen_tso.utils import ROOT






# Copyright (c) 2025 Patricio Cubillos
# Gen TSO is open-source software under the GPL-2.0 license (see LICENSE)

__all__ = [
    'get_instruments',
    'get_modes',
    'get_sed_types',
    '_spec_modes',
    '_photo_modes',
    '_acq_modes',
    '_default_aperture_strategy',
    '_get_configs',
    'get_throughputs',
    'generate_all_instruments',
    '_load_flux_rate_splines',
    'Detector',
]

from itertools import product
import pickle

import numpy as np
from scipy.interpolate import CubicSpline
from pandeia.engine.calc_utils import get_instrument_config

from ..utils import ROOT


# Dictionaries
inst_names = {
    'miri': 'MIRI',
    'nircam': 'NIRCam',
    'niriss': 'NIRISS',
    'nirspec': 'NIRSpec',
}

spec_dict = {
    'miri': ['lrsslitless', 'lrsslit', 'mrs_ts'],
    'nircam': ['lw_tsgrism', 'sw_tsgrism'],
    'niriss': ['soss'],
    'nirspec': ['bots'],
}

photo_dict = {
    'miri': ['imaging_ts'],
    'nircam': ['lw_ts', 'sw_ts'],
}

# Lists
_spec_modes = []
for modes in spec_dict.values():
    _spec_modes += modes

_photo_modes = []
for modes in photo_dict.values():
    _photo_modes += modes

_acq_modes = [
    'target_acq',
]

def get_instruments():
    """
    Get the list of JWST instruments.

    Returns
    -------
    instruments: 1D list of strings
        JWST instruments
    """
    return list(inst_names)


def get_modes(instrument, type=None):
    """
    Get the list of available TSO modes for a given instrument.

    Parameters
    ----------
    instrument: String
        A JWST instrument.
    type: String
        If set, get only the modes of the specified type:
        spectroscopy, photometry, or acquisition

    Returns
    -------
    modes: 1D list of strings
        JWST TSO modes for instrument
    """
    modes = []
    if type is None:
        types = ['spectroscopy', 'photometry', 'acquisition']
    else:
        types = [type]

    if 'spectroscopy' in types:
        modes += spec_dict[instrument]
    if 'photometry' in types and instrument in photo_dict:
        modes += photo_dict[instrument]
    if 'acquisition' in types:
        modes += _acq_modes

    return modes


def get_sed_types():
    """
    Get the list of SED models.

    Returns
    -------
    instruments: 1D list of strings
        JWST instruments
    """
    return [
        'phoenix',
        'k93models',
        'bt_settl',
    ]


# Spectral extraction apertures (arcsec) based on values reported in:
# Ahrer et al. (2023)    NIRCam/LW
# Alderson et al. (2023) NIRSpec/G395H
# Bouwman et al. (2024)  MIRI/LRS
# Bell et al. (2024)     MIRI/LRS
_default_aperture_strategy = {
    'lrsslitless': dict(
        aperture_size = 0.6,
        sky_annulus = [1.0, 2.5],
    ),
    'lrsslit': dict(
        aperture_size = 0.88,
        sky_annulus = [0.88, 1.4],
    ),
    'mrs_ts': dict(
        aperture_size = 0.6,
        sky_annulus = [1.0, 1.5],
    ),
    'lw_tsgrism': dict(
        aperture_size = 0.6,
        sky_annulus = [0.9, 1.5],
    ),
    'sw_tsgrism': dict(
        aperture_size = 0.6,
        sky_annulus = [0.9, 1.5],
    ),
    'bots': dict(
        aperture_size = 0.7,
        sky_annulus = [0.7, 1.5],
    ),
}


def get_constraints(config, inst_property, mode, **prop_constraints):
    """
    Extract instrument properties that might be constrained or not.

    Parameters
    ----------
    config: Dict
        Instrument config dict.
    inst_property: String
        The property to extract. Select from: dispersers, filters,
        subarrays, or readout_patterns.
    mode: String
        Observing mode.
    **prop_constraints: String
        Constraining property and value.
    """
    prop_key = list(prop_constraints)[0]
    constraint = prop_constraints[prop_key]

    constraints = {}
    if prop_key in config['config_constraints']:
        constraints = config['config_constraints'][prop_key]
    is_constrained = (
        constraint is not None and
        constraint in constraints and
        inst_property in constraints[constraint]
    )
    if is_constrained:
        constraint_val = constraints[constraint][inst_property]
        if isinstance(constraint_val, list):
            return constraint_val
        if mode in constraint_val:
            return constraint_val[mode]
        if constraint_val['default'] is None:
            return config['mode_config'][mode][inst_property]
        return constraint_val['default']
    if inst_property in config['mode_config'][mode]:
        return config['mode_config'][mode][inst_property]
    return config[inst_property]


def str_or_dict(info_dict, selected, mode):
    """
    Extract the label of a requested property.

    Parameters
    ----------
    info_dict: Dict
        Instrument's disperser, readout, filter, subarray or slit config
    selected: String
        Selected property.
    mode: String
        Observing mode.
    """
    if 'display_string' not in info_dict[selected]:
        return selected.upper()
    if isinstance(info_dict[selected]['display_string'], str):
        return info_dict[selected]['display_string']
    if mode in info_dict[selected]['display_string']:
        return info_dict[selected]['display_string'][mode]
    return info_dict[selected]['display_string']['default']


def _get_configs(instrument=None, obs_type=None):
    """
    Collect the information from the available observing modes
    (Names, modes, dispersers, readout patterns, subarrays, slits).

    Parameters
    ----------
    instrument: String or List
        The JWST instrument(s) to collect. Select one or more
        from miri, nircam, nirspec, or niriss.
        If None collect all instruments.
    obs_type: String
        Type of observation: spectroscopy or acquisition

    Returns
    -------
    A list of dictionaries containing the observing modes info.

    Some properties have special constraints which are specified
    by the 'constraints' item, which is a nested dict of the shape:
        options = constraints[constrained_field][constrainer_field][val]
    For example, for nirspec bots the filters options depend on the disperser,
    then to get the filters available for the 'g140h' disperser do:
        constrained_filters = constraints['filter']['disperser']['g140h']

    Examples
    --------
    >>> from gen_tso.pandeia_io import _get_configs

    >>> insts = _get_configs(obs_type='spectroscopy')
    >>> insts = _get_configs(obs_type='photometry')
    >>> insts = _get_configs(obs_type='acquisition')

    >>> insts = _get_configs(instrument='miri')
    >>> insts = _get_configs(instrument='nircam', obs_type='spectroscopy')
    """
    ta_apertures = {
        'miri': ['imager'],
        'nircam': ['lw'],
        'niriss': ['imager', 'nrm'],
        'nirspec': ['s1600a1'],
    }
    if instrument is None:
        instrument = 'miri nircam nirspec niriss'.split()
    elif isinstance(instrument, str):
        instrument = [instrument]

    if obs_type is None:
        obs_type = ['spectroscopy', 'acquisition']
    elif isinstance(obs_type, str):
        obs_type = [obs_type]

    modes = []
    if 'spectroscopy' in obs_type:
        modes += _spec_modes
    if 'photometry' in obs_type:
        modes += _photo_modes
    if 'acquisition' in obs_type:
        modes += _acq_modes

    telescope = 'jwst'
    outputs = []
    for mode, inst in product(modes, instrument):
        config = get_instrument_config(telescope, inst)
        if mode not in config['modes']:
            continue

        names = {
            'dispersers': config['disperser_config'],
            'filters': config['filter_config'],
            'readout_patterns': config['readout_pattern_config'],
            'subarrays': config['subarray_config']['default'],
        }
        if 'slit_config' in config:
            names['slits'] = config['slit_config']
        props = list(names)

        if mode in _spec_modes:
            obs_type = 'spectroscopy'
        elif mode in _acq_modes:
            obs_type = 'acquisition'
        elif mode in _photo_modes:
            obs_type = 'photometry'

        inst_dict = {}
        inst_dict['instrument'] = inst_names[inst]
        inst_dict['obs_type'] = obs_type
        mode_name = config['mode_config'][mode]['display_string']
        inst_dict['mode'] = mode
        inst_dict['mode_label'] = mode_name

        aperture_names = config['aperture_config']
        apertures = config['mode_config'][mode]['apertures']
        if obs_type == 'acquisition':
            apertures = [
                aper for aper in apertures
                if aper in ta_apertures[inst]
            ]
        inst_dict['apertures'] = {
            aperture: str_or_dict(aperture_names, aperture, mode)
            for aperture in apertures
        }
        aper = apertures[0] if mode=='target_acq' else None

        for prop in props:
            vals = get_constraints(config, prop, mode, apertures=aper)
            prop_name = 'readouts' if prop=='readout_patterns' else prop
            inst_dict[prop_name] = {
                value: str_or_dict(names[prop], value, mode)
                for value in vals
            }

        if inst_dict['mode'] == 'soss':
            inst_dict['orders'] = {
                '1': '1',
                '2': '2',
                '1 2': '1 and 2',
            }
        else:
            inst_dict['orders'] = None

        if obs_type == 'acquisition':
            groups = config['mode_config'][mode]['enum_ngroups']
            if not isinstance(groups, list):
                groups = groups['default']
            inst_dict['groups'] = groups
            # Integrations and exposures are always one (so far)
            #print(config['mode_config'][mode]['enum_nexps'])
            #print(config['mode_config'][mode]['enum_nints'])


        # Special constraints
        inst_dict['constraints'] = {}
        # MIRI
        if inst_dict['instrument']=='MIRI' and mode=='target_acq':
            group_constraints = config['mode_config'][mode]['enum_ngroups']
            constraints = {}
            for subarray in inst_dict['subarrays']:
                key = subarray if subarray in group_constraints else 'default'
                constraints[subarray] = group_constraints[key]
            inst_dict['constraints']['groups'] = {'subarrays': constraints}

        # NIRCam
        if mode == 'sw_tsgrism':
            constraints = {
                aper: get_constraints(config, 'subarrays', mode, apertures=aper)
                for aper in apertures
            }
            inst_dict['constraints']['subarrays'] = {'apertures': constraints}

        if mode == 'lw_tsgrism':
            constraints = {
                subarray: get_constraints(
                    config, 'readout_patterns', mode, subarrays=subarray,
                )
                for subarray in inst_dict['subarrays']
            }
            inst_dict['constraints']['readouts'] = {'subarrays': constraints}

        if inst == 'nircam' and obs_type == 'photometry':
            constraints = {
                aper: get_constraints(
                    config, 'filters', mode, apertures=aper,
                )
                for aper in apertures
            }
            double_filters = config['double_filters']

            pupil_to_aperture = {}
            pupils = {}
            filters = {}
            for aper, a_label in inst_dict['apertures'].items():
                pupils[aper] = a_label
                filters[aper] = []
                for filter in constraints[aper]:
                    if filter in double_filters:
                        filters[filter] = [filter]
                        pupils[filter] = filter.upper()
                        pupil_to_aperture[filter] = aper
                    else:
                        filters[aper] += [filter]
                pupil_to_aperture[aper] = aper
            inst_dict['constraints']['filters'] = {'pupils': filters}
            inst_dict['pupils'] = pupils
            inst_dict['pupil_to_aperture'] = pupil_to_aperture

        if mode == 'sw_ts':
            inst_dict['constraints']['filters']['apertures'] = constraints
            constraints = {
                aper: get_constraints(
                    config, 'subarrays', mode, apertures=aper,
                )
                for aper in apertures
            }
            inst_dict['constraints']['subarrays'] = {'apertures': constraints}
            pairings = {
                'imaging': 'LW Imaging Time Series',
                'grism': 'LW Grism Time Series',
            }
            constraints = {pairing: [] for pairing in pairings}
            for pupil in inst_dict['pupils']:
                if inst_dict['pupil_to_aperture'][pupil] in ['sw', 'wlp8__ts']:
                    constraints['imaging'].append(pupil)
                else:
                    constraints['grism'].append(pupil)
            inst_dict['pairings'] = pairings
            inst_dict['constraints']['pupils'] = {'pairings': constraints}

        # NIRSpec
        if mode == 'bots':
            constraints = {
                disp: get_constraints(config, 'filters', mode, dispersers=disp)
                for disp in inst_dict['dispersers']
            }
            inst_dict['constraints']['filters'] = {'dispersers': constraints}
            constraints = {}
            for disperser in inst_dict['dispersers']:
                subs = list(inst_dict['subarrays'])
                if disperser == 'prism':
                    subs.remove('sub1024a')
                constraints[disperser] = subs
            inst_dict['constraints']['subarrays'] = {'dispersers': constraints}

        # NIRISS
        if mode == 'soss':
            constraints = {
                'substrip96': ['1'],
                'substrip256': ['1', '2', '1 2'],
                'sossfull': ['1', '2', '1 2'],
            }
            inst_dict['constraints']['orders'] = {'subarrays': constraints}
            constraints = {
                'substrip96': ['nisrapid'],
                'substrip256': ['nisrapid'],
                'sossfull': inst_dict['readouts'],
            }
            inst_dict['constraints']['readouts'] = {'subarrays': constraints}

        if inst_dict['instrument']=='NIRISS' and mode=='target_acq':
            constraints = {
                aper: get_constraints(
                    config, 'readout_patterns', mode, apertures=aper,
                )
                for aper in inst_dict['apertures']
            }
            inst_dict['constraints']['readouts'] = {'apertures': constraints}

        outputs.append(inst_dict)
    return outputs


def print_configs(instrument, mode, output):
    telescope = 'jwst'
    inst_config = get_instrument_config(telescope, instrument)

    if mode is None:
        return(inst_config['modes'])
    config = inst_config['mode_config'][mode]

    subarrays = config['subarrays']
    screen_output = f'subarrays: {subarrays}\n'

    if instrument == 'niriss':
        readouts = inst_config['readout_patterns']
    else:
        readouts = config['readout_patterns']
    screen_output += f'readout patterns: {readouts}\n'

    if instrument == 'nirspec':
        gratings_dict = inst_config['config_constraints']['dispersers']
        gratings = filters = dispersers = []
        for grating, filter_list in gratings_dict.items():
            for filter in filter_list['filters']:
                gratings.append(f'{grating}/{filter}')
        screen_output += f'grating/filter pairs: {gratings}'
    else:
        filters = config['filters']
        dispersers = [disperser for disperser in config['dispersers']]
        screen_output += f'dispersers: {dispersers}\n'
        screen_output += f'filters: {filters}'

    if output is None:
        print(screen_output)
    elif output == 'readouts':
        return readouts
    elif output == 'subarrays':
        return subarrays
    elif output == 'filters':
        return filters
    elif output == 'dispersers':
        return dispersers
    elif instrument=='nirspec' and output=='gratings':
        return gratings
    else:
        raise ValueError(f"Invalid config output: '{output}'")


class Detector:
    def __init__(
            self, mode, label, instrument, obs_type,
            aperture_label, apertures,
            disperser_label, dispersers,
            filter_label, filters,
            subarrays, readouts, orders=None,
            groups=None, default_indices=None,
            constraints={},
        ):
        """
        An object containing the available instrumental setup options.
        (Intended for the front-end)

        The constraints are a triple-nested dict of the form:
        constraints[constrained_var][constraining_var][var] = constraints
        For example, readout values constrained by the disperser:
        constraints['readouts']['dispersers'] = {
            'imager': ['nis', 'nisrapid'],
            'nrm': ['nisrapid'],
        }
        """
        self.mode = mode
        self.mode_label = label
        self.instrument = instrument
        self.obs_type = obs_type
        self.aperture_label = aperture_label
        self.apertures = apertures
        self.disperser_label = disperser_label
        self.dispersers = dispersers
        self.filter_label = filter_label
        self.filters = filters
        self.subarrays = subarrays
        self.readouts = readouts
        self.groups = groups
        self.constraints = constraints
        if self.mode == 'soss':
            self.orders = orders

        telescope = 'jwst'
        self.ins_config = get_instrument_config(telescope, instrument.lower())

        if default_indices is None:
            idx_a = idx_d = idx_f = idx_s = idx_r = 0
        else:
            idx_a, idx_d, idx_f, idx_s, idx_r = default_indices
        self.default_aperture = list(apertures)[idx_a]
        if self.mode == 'target_acq':
            self.default_disperser = None
        else:
            self.default_disperser = list(dispersers)[idx_d]
        self.default_filter = list(filters)[idx_f]
        self.default_subarray = list(subarrays)[idx_s]
        self.default_readout = list(readouts)[idx_r]

    def get_constrained_val(
            self, var,
            disperser=None, filter=None, subarray=None, readout=None,
            aperture=None, pupil=None, pairing=None,
        ):
        """
        det = generate_all_instruments()[5]
        det.get_constrained_val('filters', disperser='nrm')
        get_constrained_val(det, 'readouts', disperser='nrm')
        get_constrained_val(det, 'readouts', disperser='imager')
        get_constrained_val(det, 'readouts', filter='imager')
        get_constrained_val(det, 'readouts')
        get_constrained_val(det, 'readouts', disperser='lala')
        """
        choices = {
            'dispersers': disperser,
            'filters': filter,
            'subarrays': subarray,
            'readouts': readout,
            'apertures': aperture,
            'pupils': pupil,
            'pairings': pairing,
        }

        default_vals = getattr(self, var)
        if var not in self.constraints:
            return default_vals

        for field, constraint in choices.items():
            if constraint is None:
                continue
            is_constrained = (
                field in self.constraints[var] and
                constraint in self.constraints[var][field]
            )
            if is_constrained:
                values = self.constraints[var][field][constraint]
                if isinstance(default_vals, list):
                    return values
                return {
                    val:label
                    for val,label in default_vals.items()
                    if val in values
                }
        return default_vals

    def instrument_label(self, disperser, filter):
        """
        Generate a pretty label with only instrument, mode,
        and sometimes disperser or filter.

        Examples
        --------
        >>> import gen_tso.pandeia_io as jwst
        >>> detectors = jwst.generate_all_instruments()
        >>> bots = detectors[5]
        >>> print(bots.instrument_label('g395h', 'f290lp'))
        """
        inst = self.instrument
        mode = self.mode
        if mode == 'mrs_ts':
            disperser_label = self.dispersers[disperser]
            label = f'{inst} / MRS / {disperser_label}'
        elif mode in ['lrsslit', 'lrsslitless']:
            label = f'{inst} / LRS / {mode[3:].upper()}'
        elif mode == 'imaging_ts':
            filter_label = self.filters[filter]
            label = f'{inst} / {filter_label}'

        elif mode == 'soss':
            label = f'{inst} / SOSS'

        elif mode == 'lw_tsgrism':
            filter_label = self.filters[filter]
            label = f'{inst} / {filter_label}'
        elif mode == 'sw_tsgrism':
            filter_label = self.filters[filter]
            label = f'{inst} / {filter_label}'
        elif mode == 'lw_ts' or mode == 'sw_ts':
            filter_label = self.filters[filter]
            label = f'{inst} / {filter_label}'

        elif mode == 'bots':
            label = self.filters[f'{disperser}/{filter}']
            disperser_label = label[:label.index('/')]
            label = f'{inst} / {disperser_label}'

        elif mode == 'target_acq':
            label = f'{inst} / acquisition'

        return label


def get_throughputs(type=None, inst=None, mode=None):
    """
    Collect the throughput response curves for each instrument configuration

    Parameters
    ----------
    type: String
        If set, get only the modes of the specified type:
        spectroscopy, photometry, or acquisition
    instrument: String
        If set, get only throughputs for the specified instrument.
    mode: String
        If set, get only throughputs for the specified mode and instrument.

    Returns
    -------
    throughputs: Dictionary
        Nested dictionary of shape throughputs[type][inst][mode][key][filter]
        containing the wavelength (um) and response throughputs.
        'key' refers to the apertures for photometry modes, and subarrays
        for others.

    Examples
    --------
    >>> import gen_tso.pandeia_io as jwst
    >>>
    >>> # All available throughputs
    >>> throughputs = jwst.get_throughputs()
    >>> # All throughputs for NIRSpec
    >>> throughputs = jwst.get_throughputs(inst='nirspec')
    """
    if mode is not None:
        if inst is None:
            msg = 'Also need to specify an instrument when requesting a specific mode'
            raise ValueError(msg)
        t_file = f'{ROOT}data/throughputs/throughputs_{inst}_{mode}.pickle'
        with open(t_file, 'rb') as handle:
            throughputs = pickle.load(handle)
        return throughputs

    if inst is not None:
        modes = get_modes(inst)
        throughputs = {}
        for mode in modes:
            t_file = f'{ROOT}data/throughputs/throughputs_{inst}_{mode}.pickle'
            with open(t_file, 'rb') as handle:
                data = pickle.load(handle)
            throughputs[mode] = data
        return throughputs

    if type is None:
        types = ['spectroscopy', 'photometry', 'acquisition']
    else:
        types = [type]
    instruments = get_instruments()

    throughputs = {}
    for obs_type,inst in product(types, instruments):
        modes = get_modes(inst, obs_type)
        if obs_type not in types or len(modes) == 0:
            continue
        if obs_type not in throughputs:
            throughputs[obs_type] = {}
        throughputs[obs_type][inst] = {}

        for mode in modes:
            t_file = f'{ROOT}data/throughputs/throughputs_{inst}_{mode}.pickle'
            with open(t_file, 'rb') as handle:
                data = pickle.load(handle)
            throughputs[obs_type][inst][mode] = data

    if type is not None:
        return throughputs[type]
    return throughputs


def generate_all_instruments():
    """
    A list of Detector() objects to keep the instrument observing mode
    configurations.

    Examples
    --------
    >>> from gen_tso.pandeia_io import _get_configs, generate_all_instruments
    >>> spec_insts = _get_configs(obs_type='spectroscopy')
    >>> photo_insts = _get_configs(obs_type='photometry')
    >>> acq_insts = _get_configs(obs_type='acquisition')
    >>> dets = generate_all_instruments()
    >>> det = dets[5]
    """
    detectors = []
    # Spectroscopic observing modes
    spec_insts = _get_configs(obs_type='spectroscopy')
    for inst in spec_insts:
        mode = inst['mode']
        dispersers = inst['dispersers']
        filters = inst['filters']
        subarrays = inst['subarrays']
        readouts = inst['readouts']
        apertures = inst['apertures']
        orders = inst['orders']
        constraints = inst['constraints']

        aperture_label = 'Aperture'
        if mode in ['lrsslit', 'lrsslitless']:
            disperser_label = 'Disperser'
            filter_label = ''
            filters = {'': ''}
            default_indices = 0, 0, 0, 0, 0
        if mode == 'mrs_ts':
            disperser_label = 'Wavelength Range'
            filter_label = ''
            filters = {'': ''}
            default_indices = 0, 0, 0, 0, 0
        if mode == 'lw_tsgrism':
            disperser_label = 'Grism'
            filter_label = 'Filter'
            default_indices = 0, 0, 3, 3, 0
        if mode == 'sw_tsgrism':
            aperture_label = 'PSF Type'
            disperser_label = 'Grism'
            filter_label = 'Filter'
            default_indices = 3, 0, 4, 0, 0
        if mode == 'bots':
            aperture_label = 'Slit'
            disperser_label = 'Grating'
            filter_label = 'Grating/Filter'
            inst['mode_label'] = 'Bright Object Time Series (BOTS)'
            filter_constraints = constraints['filters']['dispersers']
            gratings = {}
            for filter in filters:
                for disperser, c_filters in filter_constraints.items():
                    if filter in c_filters:
                        label = f"{dispersers[disperser]}/{filters[filter]}"
                        gratings[f'{disperser}/{filter}'] = label
            filters = gratings
            default_indices = 0, 0, 7, 4, 1
        if mode == 'soss':
            disperser_label = 'Disperser'
            filter_label = 'Filter'
            inst['mode_label'] = 'Single Object Slitless Spectroscopy (SOSS)'
            default_indices = 0, 0, 0, 0, 1

        det = Detector(
            mode,
            inst['mode_label'],
            inst['instrument'],
            inst['obs_type'],
            aperture_label,
            apertures,
            disperser_label,
            dispersers,
            filter_label,
            filters,
            subarrays,
            readouts,
            orders=orders,
            default_indices=default_indices,
            constraints=constraints,
        )
        detectors.append(det)

    # Photometry observing modes
    photo_insts = _get_configs(obs_type='photometry')
    for inst in photo_insts:
        mode = inst['mode']
        apertures = inst['apertures']
        dispersers = inst['dispersers']
        filters = inst['filters']
        subarrays = inst['subarrays']
        readouts = inst['readouts']
        constraints = inst['constraints']

        filter_label = 'Filter'
        if mode == 'lw_ts':
            aperture_label = 'LW Pupil'
            filter_label = 'LW Filter'
        elif mode == 'sw_ts':
            aperture_label = 'SW Pupil'
            filter_label = 'SW Filter'
        disperser_label = 'Disperser'
        dispersers = {'':''}
        default_indices = 0, 0, 0, 0, 0

        det = Detector(
            mode,
            inst['mode_label'],
            inst['instrument'],
            inst['obs_type'],
            aperture_label,
            apertures,
            disperser_label,
            dispersers,
            filter_label,
            filters,
            subarrays,
            readouts,
            default_indices=default_indices,
            constraints=constraints,
        )
        detectors.append(det)

        if mode in ['lw_ts', 'sw_ts']:
            det.pupils = inst['pupils']
            det.pupil_to_aperture = inst['pupil_to_aperture']
        if mode == 'sw_ts':
            det.pairings = inst['pairings']

    # Acquisition observing modes
    acq_insts = _get_configs(obs_type='acquisition')
    for inst in acq_insts:
        mode = inst['mode']
        aperture_label = 'Acquisition mode'
        apertures = inst['apertures']
        disperser_label = 'Disperser'
        dispersers = inst['dispersers']
        dispersers = {'':''}
        filter_label = 'Filter'
        filters = inst['filters']
        subarrays = inst['subarrays']
        readouts = inst['readouts']
        constraints = inst['constraints']

        if inst['instrument'] == 'MIRI':
            default_indices = 0, 0, 0, 5, 0
        if inst['instrument'] == 'NIRCam':
            default_indices = 0, 0, 0, 0, 0
        if inst['instrument'] == 'NIRISS':
            default_indices = 0, 0, 0, 0, 1
        if inst['instrument'] == 'NIRSpec':
            default_indices = 0, 0, 0, 1, 0

        det = Detector(
            mode,
            inst['mode_label'],
            inst['instrument'],
            inst['obs_type'],
            aperture_label,
            apertures,
            disperser_label,
            dispersers,
            filter_label,
            filters,
            subarrays,
            readouts,
            groups=inst['groups'],
            default_indices=default_indices,
            constraints=constraints,
        )
        detectors.append(det)

    return detectors


def get_detector(instrument=None, mode=None, detectors=None):
    """
    Find the detector matching instrument and mode.
    """
    if detectors is None:
        detectors = generate_all_instruments()

    if instrument is None and mode is None:
        return None

    if mode is None:
        # Default to first detector for instrument (spectroscopic mode)
        for detector in detectors:
            if detector.instrument.lower() == instrument.lower():
                return detector

    for det in detectors:
        if det.mode == mode:
            if instrument is None or det.instrument.lower()==instrument.lower():
                return det
    return None


def make_save_label(
        target, inst, mode, aperture, disperser, filter,
        subarray=None, readout=None, order=None,
    ):
    """
    A nice short name when saving to file via the application
    """
    if target is not None:
        target = '_' + target.replace(' ', '')

    if mode == 'target_acq':
        return f'tso{target}_{inst}_{mode}.pickle'
    # Shorter miri and nircam mode names:
    mode = mode.replace('mrs_ts', 'mrs')
    mode = mode.replace('w_tsgrism', 'w_grism')
    mode = mode.replace('w_ts', 'w_imaging')

    if mode == 'mrs':
        return f'tso{target}_{inst}_{mode}_{aperture}.pickle'
    elif mode == 'imaging_ts':
        return f'tso{target}_{inst}_{mode}_{filter}.pickle'
    elif mode in ['lrsslit', 'lrsslitless', 'soss']:
        return f'tso{target}_{inst}_{mode}.pickle'
    elif mode == 'bots':
        return f'tso{target}_{inst}_{mode}_{disperser}.pickle'
    elif mode == 'lw_grism':
        return f'tso{target}_{inst}_{mode}_{filter}.pickle'
    elif mode == 'sw_grism':
        return f'tso{target}_{inst}_{mode}_{subarray}_{filter}.pickle'


def make_detector_label(
        inst, mode, aperture, disperser, filter, subarray, readout, order,
    ):
    """
    Generate a pretty and (as succinct as possible) label for the
    detector configuration.
    """
    if mode == 'target_acq':
        if inst == 'miri':
            label = f'{filter.upper()} {subarray.upper()}'
        if inst == 'nircam':
            label = f'{filter.upper()} {readout.upper()}'
        if inst == 'niriss':
            label = 'Bright' if aperture == 'nrm' else 'Faint'
            label = f'{label} {readout.upper()}'
        if inst == 'nirspec':
            label = f'{filter.upper()} {subarray.upper()} {readout.upper()}'
        return f'{inst_names[inst]} {label}'

    if mode == 'mrs_ts':
        return f'MIRI MRS {disperser.upper()}'
    if mode in ['lrsslit', 'lrsslitless']:
        return f'MIRI {mode.upper()}'
    if mode == 'imaging_ts':
        return f'MIRI {filter.upper()}'

    if mode == 'soss':
        order = f' O{order[0]}' if len(order)==1 else ''
        return f'NIRISS {mode.upper()} {subarray}{order}'

    if mode == 'lw_tsgrism':
        subarray = subarray.replace('grism', '').replace('_dhs', '')
        return f'NIRCam {filter.upper()} {subarray} {readout}'
    if mode == 'sw_tsgrism':
        return f'NIRCam {filter.upper()} {subarray} {readout}'
    if mode == 'lw_ts':
        subarray = subarray.replace('grism', '').replace('_dhs', '')
        return f'NIRCam imaging {filter.upper()} {subarray} {readout}'
    if mode == 'sw_ts':
        return f'NIRCam imaging {aperture} {filter.upper()} {subarray} {readout}'

    if mode == 'bots':
        if filter == 'f070lp':
            disperser = f'{disperser}/{filter}'
        return f'NIRSpec {disperser.upper()} {subarray} {readout}'


def make_saturation_label(
        mode, aperture, disperser, filter, subarray,
        order='', sed_label='',
    ):
    """
    Make a label of unique saturation setups to identify when and
    when not the saturation level can be estimated.
    """
    if mode == 'bots':
        sat_label = f'_{disperser}_{subarray}'
    elif mode == 'soss':
        order = f'_O{order[0]}' if len(order)==1 else ''
        sat_label = f'{order}'
    elif mode == 'sw_tsgrism':
        sat_label = f'_{aperture}_{subarray}'
    elif mode == 'sw_ts':
        sat_label = f'_{aperture}'
    elif mode == 'mrs_ts':
        sat_label = f'_{disperser}'
    elif mode == 'target_acq':
        sat_label = f'_{aperture}'
    else:
         sat_label = ''

    sat_label = f'{mode}_{filter}{sat_label}_{sed_label}'
    return sat_label


def make_obs_label(
        inst, mode, aperture, disperser, filter, subarray, readout, order,
        ngroup, nint, run_type, sed_label, depth_label,
    ):
    detector_label = make_detector_label(
        inst, mode, aperture, disperser, filter, subarray, readout, order,
    )
    if run_type == 'Acquisition':
        group_ints = f'({ngroup} G)'
        depth_label = 'acquisition'
    else:
        group_ints = f'({ngroup} G, {nint} I)'

    label = (
        f'{detector_label} {group_ints} / {sed_label} / {depth_label}'
    )
    return label


def _load_flux_rate_splines(obs_label=None):
    """
    Get dictionary of cubic spline functions for pre-calculated
    brightest pixel flux rates at given instrumental and SED config
    combinations.

    Parameters
    ----------
    obs_label: String
         If not None, only return the flux_rate and full_well for the
         instrument and SED configuration matching the input label.
         Return (None,None) if no config matched the obs_label.

    Returns
    -------
    flux_rates: Dictionary of CubicSpline objects
        Spline function that returns the flux rate at a given Ks magnitude.
    full_wells: Dictionary of floats
        e- counts to saturate the detector.

    Examples
    --------
    >>> import gen_tso.pandeia_io as jwst
    >>>
    >>> # Extract all splines, evaluate for one config:
    >>> flux_rate_splines, full_wells = jwst._load_flux_rate_splines()
    >>> obs_label = 'lw_tsgrism_f444w_phoenix_k5v'
    >>> spline = flux_rate_splines[obs_label]
    >>> print(spline(8.351), full_wells[obs_label])
    3.1140479065012263 58100.0
    >>>
    >>> # Extract a single config:
    >>> obs_label = 'lw_tsgrism_f444w_phoenix_k5v'
    >>> flux_rate_spline, full_well = jwst._load_flux_rate_splines(obs_label)
    >>> print(flux_rate_spline(8.351), full_well)
    3.1140479065012263 58100.0
    """
    if obs_label is not None:
        tokens = obs_label.split('_')
        i_label = '_'.join(tokens[:-2]) + '_'

    flux_rate_data = {}
    for instrument in ['miri', 'nircam', 'niriss', 'nirspec']:
        inst = instrument.lower()
        rate_file = f'{ROOT}data/flux_rates_{inst}.pickle'
        with open(rate_file, 'rb') as handle:
            rates = pickle.load(handle)
        mag = rates.pop('magnitude')
        flux_rate_data[inst] = rates

    aper_modes = _photo_modes + ['sw_tsgrism', 'mrs_ts']

    flux_rates = {}
    full_wells = {}
    for inst, mode_rates in flux_rate_data.items():
        for mode, disp_rates in mode_rates.items():
            orders = ['1', '1 2'] if mode=='soss' else ['1']
            for disperser, filter_rates in disp_rates.items():
                for filter, sub_rates in filter_rates.items():
                    for subarray, sed_rates in sub_rates.items():
                        for order in orders:
                            if mode in aper_modes:
                                aperture = disperser
                            else:
                                aperture = ''
                            filter = filter.replace('None', '')
                            inst_label = make_saturation_label(
                                mode, aperture, disperser, filter,
                                subarray, order, ''
                            )
                            if obs_label is not None and inst_label != i_label:
                                continue
                            for sed_type in get_sed_types():
                                if sed_type not in sed_rates:
                                    continue
                                name = f'{sed_type}_names'
                                for i,rate in enumerate(sed_rates[sed_type]):
                                    log_rate = np.log10(rate)
                                    sed = sed_rates[name][i]
                                    label = f'{inst_label}{sed_type}_{sed}'
                                    if obs_label is None:
                                        flux_rates[label] = CubicSpline(mag, log_rate)
                                        full_wells[label] = sed_rates['full_well']
                                    elif label == obs_label:
                                        return CubicSpline(mag, log_rate), sed_rates['full_well']

    # Model not found
    if obs_label is not None:
        return None, None
    return flux_rates, full_wells


